/* Auspice functions.
 *
 * Auspice Services is copyright (c) 2000-2001 In Mean.
 *     E-mail: <auspice@auspice.org>
 *
 * This program is free but copyrighted software; see the file LICENSE for
 * details.
 */

#include "../inc/services.h"
#include "../inc/timeout.h"
#include "../inc/auspice.h"
#include "bs-help.c"

int servbot = 0;
int chtml = 0;
int wgameline = 0;
int bwsize = 0;

static int nhostline = 0;
int nmessage = 0;
int nhj = 0;
static int hostline_size = 0;
static int message_size = 0;
static int hj_size = 0;

struct message_ {
    char *text;
    short type;
    char *who;
    long reserved[4];
} *lmessages = NULL;

struct hostjoin_ {
    char *host;
    char *chan;
    char *who;
    long reserved[4];
} *hostjoin = NULL;

static struct s_hline {
	char *nick;
	char *host;
	char who[NICKMAX];
	long reserved[4];
} *s_hlines = NULL;

char *botname[64];
char *botmaster[64];

char *channelname[64];
char *channeltopic[64];
char *wordgame[64];
char *wordhelp[64];
char *badword[64];

int majdata;

static void do_agenthelp(const char *source);
static void do_agentset(const char *source);

int del_hostline(const char *nick);
#ifdef DEFHOSTSERV
void load_hline_db(void);
void save_hline_db(void);
#endif
void do_hostline(const char *nick);
void agent(const char *source, char *buf);
void servicesX(const char *source, char *buf);
void servicesW(const char *source, char *buf);
void do_selfdel(const char *source);

static BotInfo *makebot(char *nick);
static void insert_bot(BotInfo *bi);

int check_ifbot(char *abotname);
int is_servbot(char *nick);
static void abotcom(const char *source, ChannelInfo *ci, char *buf);
void botkick(char *nick, ChannelInfo *ci, char *buf);
void dropmode(char *nick, ChannelInfo *ci);
int is_badwords(char *text);

int socket_connected = 0;

/* 1 = connected
   0 = not connected, no failed attempts
   -(x), x = number of failed attempts */

int socket_fd = 0;
long int last_ping;

/************************************************************************/
BotInfo *botlists[256];
int nbots = 0;
/************************************************************************/
static BotInfo *makebot(char *nick)
{
    BotInfo *bi;

    bi = scalloc(sizeof(BotInfo), 1);
    bi->nick = sstrdup(nick);
    insert_bot(bi);
    nbots++;
    return bi;
}
/************************************************************************/
static void insert_bot(BotInfo *bi)
{
    /* bi->prev = NULL;
    bi->next = botlist;
    if (bi->next) bi->next->prev = bi;
    botlist = bi; */

    BotInfo *ptr, *prev;

    for (prev = NULL, ptr = botlists[tolower(*bi->nick)];
                        ptr != NULL && stricmp(ptr->nick, bi->nick) < 0;
                        prev = ptr, ptr = ptr->next)
        ;
        bi->prev = prev;
    bi->next = ptr;
    if (!prev)
        botlists[tolower(*bi->nick)] = bi;
    else
        prev->next = bi;
    if (ptr)
        ptr->prev = bi;
}
/***************************************************************************/
static int delbot(BotInfo *bi)
{
    do_expireabot(bi->nick);
    if (bi->next)
        bi->next->prev = bi->prev;
    if (bi->prev)
        bi->prev->next = bi->next;
    else
    botlists[tolower(*bi->nick)] = bi->next;

    nbots--;

    free(bi->nick);
    free(bi->user);
    free(bi->host);
    free(bi->real);
    free(bi->bmaster);

    free(bi);

    return 1;
}
/***************************************************************************/
BotInfo *findbot(char *nick)
{
    BotInfo *bi;

    if (!nick || !*nick) return NULL;

    for (bi = botlists[tolower(*nick)]; bi; bi = bi->next)
        if (!stricmp(nick, bi->nick)) return bi;

    return NULL;
}
/***************************************************************************/
void load_bs_dbase(void)
{
    FILE *f;
    int ver, i, j, len;
    BotInfo *bi;

    if (!(f = open_db(s_NickServ, BOTSERV_DB, "r", BS_VERSION)))
        return;

    switch (ver = get_file_version(f, BOTSERV_DB)) {
      case 1:
            while (fgetc(f) != 0) {
		char *s;
                bi = smalloc(sizeof(BotInfo));
                if (1 != fread(bi, sizeof(BotInfo), 1, f))
                       fatal_perror("Read error on %s", BOTSERV_DB);
		s = read_string(f, BOTSERV_DB);
		bi = makebot(s);
                if (bi->user)
                    bi->user = read_string(f, BOTSERV_DB);
                if (bi->host)
                    bi->host = read_string(f, BOTSERV_DB);
                if (bi->real)
                    bi->real = read_string(f, BOTSERV_DB);
                if (bi->bmaster)
                    bi->bmaster = read_string(f, BOTSERV_DB);
            } /* while (fgetc(f) == 1) */
        break;
      default:
        fatal("Unsupported version number (%d) on %s", i, BOTSERV_DB);
    } /* switch (version) */

    close_db(f, BOTSERV_DB);
}
/***************************************************************************/
/*
void save_bs_dbase(void)
{
    FILE *f;
    int i, j, len;
    BotInfo *bi;

    if (!(f = open_db(s_BotServ, BOTSERV_DB, "w", BS_VERSION)))
        return;

    for (i = 33; i < 256; ++i) {
	for (bi = botlists[i]; bi; bi = bi->next)
            fputc(1, f);
            if (1 != fwrite(bi, sizeof(BotInfo), 1, f))
            fatal_perror("Write error on %s", BOTSERV_DB);

            if (bi->nick)
                write_string(bi->nick, f, BOTSERV_DB);
	    
            if (bi->user) 
                write_string(bi->user, f, BOTSERV_DB);
	    
            if (bi->host) 
                write_string(bi->host, f, BOTSERV_DB);
	    
            if (bi->real) 
                write_string(bi->real, f, BOTSERV_DB);
	    
            if (bi->bmaster) 
                write_string(bi->bmaster, f, BOTSERV_DB);
	    
        }
        fputc(0, f);
    }
    close_db(f, BOTSERV_DB);
}
*/
/***************************************************************************/
int  is_validnick(char *nick)
{
        char *ch;

        if (*nick == '-' || isdigit(*nick))    /* first character in  [0..9-] */
                return 0;

	if (strlen(nick) > 20)
                return 0;

        for (ch = nick; *ch && (ch - nick) < 30; ch++)
                if (isspace(*ch) || (*ch == '%'))
	                return 0;

        return 1;
}
/*************************** TOOLS *************************************/
char timestamp_buff[50];
char *timestamp(void) {
        struct  tm      tm;
        time_t  timer;

        time(&timer);
        tm = *localtime(&timer); strftime(timestamp_buff, sizeof(timestamp_buff) - 1, "[%H:%M:%S %m/%d/%Y]", &tm);
        return(timestamp_buff);
}

/*************************************************************************/
void load_msg_db(void)
{
    FILE *f;
    int i;

    if (!(f = open_db(s_Agent, MSG_DB, "r", 1)))
        return;

    switch (i = get_file_version(f, MSG_DB)) {
      case 6:
      case 5:
        nmessage = fgetc(f) * 256 + fgetc(f);
        if (nmessage < 8)
            message_size = 16;
        else
            message_size = 2*nmessage;
        lmessages = smalloc(sizeof(*lmessages) * message_size);
        if (!nmessage) {
            fclose(f);
            return;
        }
        if (nmessage != fread(lmessages, sizeof(*lmessages), nmessage, f))
            fatal_perror("Read error on %s", MSG_DB);
        for (i = 0; i < nmessage; i++) {
            lmessages[i].text = read_string(f, MSG_DB);
            lmessages[i].who = read_string(f, MSG_DB);
        }
        break;
      default:
        fatal("Unsupported version (%d) on %s", i, MSG_DB);
    } /* switch (version) */

    close_db(f, MSG_DB);
}
/*************************************************************************/
void save_message(void)
{
    FILE *f;
    int i, count = 0;
    f = open_db(s_Agent, MSG_DB, "w", AS_MSG_VERSION);
    fputc(nmessage/256, f);
    fputc(nmessage & 255, f);
    if (fwrite(lmessages, sizeof(*lmessages), nmessage, f) != nmessage)
        fatal_perror("Write error on %s", MSG_DB);
    for (i = 0; i < nmessage; i++) {
        write_string(lmessages[i].text, f, MSG_DB);
        write_string(lmessages[i].who, f, MSG_DB);
    }
    close_db(f, MSG_DB);
}
/*************************************************************************/
void do_logonmsg(const char *source)
{
	int i;
        for (i = 0; i < nmessage; ++i)
            if (lmessages[i].type == M_LOGON)
                notice(s_GlobalNoticer, source, "%s", lmessages[i].text);
}
/*************************************************************************/
void do_operlogonmsg(const char *source)
{
	int i;
        for (i = 0; i < nmessage; ++i)
            if (lmessages[i].type == M_OPER)
 send_cmd(server_name, "372 %s :\00312Oper\0034News\0031: %s", source, lmessages[i].text);
}
/*************************************************************************/
#ifdef DEFHOSTSERV
void load_hline_db(void)
{
    FILE *f;
    int i;

    if (!(f = open_db(s_ServicesW, HLINE_DB, "r", 1)))
        return;

    switch (i = get_file_version(f, HLINE_DB)) {
      case 6:
      case 5:
        nhostline = fgetc(f) * 256 + fgetc(f);
        if (nhostline < 8)
            hostline_size = 16;
        else
            hostline_size = 2*nhostline;
        s_hlines = smalloc(sizeof(*s_hlines) * hostline_size);
        if (!nhostline) {
            fclose(f);
            return;
        }
        if (nhostline != fread(s_hlines, sizeof(*s_hlines), nhostline, f))
            fatal_perror("Read error on %s", HLINE_DB);
        for (i = 0; i < nhostline; i++) {
            s_hlines[i].nick = read_string(f, HLINE_DB);
            s_hlines[i].host = read_string(f, HLINE_DB);
        }
        break;
      default:
        fatal("Unsupported version (%d) on %s", i, HLINE_DB);
    } /* switch (version) */

    close_db(f, HLINE_DB);
}
#endif
/***********************************************************/
void load_majoin_db(void)
{
    FILE *f;
    int i;

    if (!(f = open_db(s_Agent, HJ_DB, "r", 1)))
        return;

    switch (i = get_file_version(f, HJ_DB)) {
      case 6:
      case 5:
        nhj = fgetc(f) * 256 + fgetc(f);
        if (nhj < 8)
            hj_size = 16;
        else
            hj_size = 2*nhj;
        hostjoin = smalloc(sizeof(*hostjoin) * hj_size);
        if (!nhj) {
            fclose(f);
            return;
        }
        if (nhj != fread(hostjoin, sizeof(*hostjoin), nhj, f))
            fatal_perror("Read error on %s", HJ_DB);
        for (i = 0; i < nhj; i++) {
            hostjoin[i].host = read_string(f, HJ_DB);
            hostjoin[i].chan = read_string(f, HJ_DB);
            hostjoin[i].who = read_string(f, HJ_DB);
        }
        break;
      default:
        fatal("Unsupported version (%d) on %s", i, HJ_DB);
    } /* switch (version) */

    close_db(f, HJ_DB);
}
/**********************************************************/
void save_majoin_db(void)
{
    FILE *f;
    int i, count = 0;
    f = open_db(s_Agent, HJ_DB, "w", AS_MJ_VERSION);
    fputc(nhj/256, f);
    fputc(nhj & 255, f);
    if (fwrite(hostjoin, sizeof(*hostjoin), nhj, f) != nhj)
        fatal_perror("Write error on %s", HJ_DB);
    for (i = 0; i < nhj; i++) {
        write_string(hostjoin[i].host, f, HJ_DB);
        write_string(hostjoin[i].chan, f, HJ_DB);
        write_string(hostjoin[i].who, f, HJ_DB);
    }
    close_db(f, HJ_DB);
}
/***********************************************************/
void load_chtml_db(void)
{
    FILE *f;
    char chtmldata[128];
    int i = 1;
    *chtmldata = 0;
    chtml = 1;
    strcat(chtmldata, "chtml.conf");
    if (f = fopen(chtmldata, "r")) {
    while(fgets(chtmldata, sizeof(chtmldata), f)) {
//	log(chtmldata);
	 if (*chtmldata == '@') {
			channelname[chtml] = sstrdup(strtok(chtmldata, "\n"));
			channeltopic[chtml] = NULL;
			chtml++;
			i = 1;
	 } else {
		 if (chtmldata && (strlen(chtmldata) > 1))
		 if (i==1) {
			channelname[chtml] = sstrdup(strtok(chtmldata, "\n"));
			i = 2;
		 } else {
			channeltopic[chtml] = sstrdup(strtok(chtmldata, "\n"));
			i = 1;
			chtml++;
		 }
	 }
    }
    fclose(f);
    }
    chtml--;
}
/***********************************************************/
void load_wordgame_db(void)
{
    FILE *f;
    char gworddata[128];
    int j;
    *gworddata = 0;
    wgameline = 1;
    if (f = fopen("wgame.conf", "r")) {
    while(fgets(gworddata, sizeof(gworddata), f)) {
	j = strlen(gworddata);
	wordgame[wgameline] = sstrdup(strtok(gworddata, " "));
	if (strlen(wordgame[wgameline]) >= j) {
		wordhelp[wgameline] = sstrdup("No Help");
	} else {
		wordhelp[wgameline] = sstrdup(strtok(NULL, "\n"));
	}
	wordgame[wgameline] = sstrdup(strtok(wordgame[wgameline], "\n"));
	wgameline++;
    }
    fclose(f);
    }
    wgameline--;
}
/***********************************************************/
void load_nsmsg_db(void)
{
    FILE *f;
    char gworddata[128];
    int j;
    *gworddata = 0;
    nsmsgcount = 1;
    if (f = fopen("nsmess.conf", "r")) {
	    while(fgets(gworddata, sizeof(gworddata), f)) {
//		log("1: %s", gworddata);
		if ((strlen(gworddata) > 1) && (*gworddata != '#')) {
			nsmsg[nsmsgcount] = sstrdup(strtok(gworddata, "\n"));
			nsmsgcount++;
		}
	    }
	    fclose(f);
    }
    nsmsgcount--;
}
/*******************************************************/
int is_badwords(char *text)
{
	int i;
	if (!text)
		return 0;

	for (i=1; i<=bwsize; i++) {
		if(strchr(badword[i], '*')) {
			if (do_match_wild(badword[i], text, 0))
				return 1;
		} else {
			if (stristr(text, badword[i]))
				return 1;
		}
	}
	return 0;
}
/*******************************************************/
void load_bw_db(void)
{
    FILE *f;
    char gworddata[128];
    int j;
    *gworddata = 0;
    bwsize = 1;
    if (f = fopen("badword.conf", "r")) {
	    while(fgets(gworddata, sizeof(gworddata), f)) {
		if ((strlen(gworddata) > 3) && (*gworddata != '#')) {
			badword[bwsize] = sstrdup(strtok(gworddata,"\n"));
			bwsize++;
		}
	    }
	    fclose(f);
    }
    bwsize--;
}
/***********************************************************/
void make_channel_html(void)
{
    Channel *c;
    int i;
    FILE *fp;
    FILE *fp1;

    if (ishtml == 0)
	return;

    fp = fopen(html_path, "w");
    fp1 = fopen("channel.dump", "w");
    if (fp) {
	fprintf(fp, "<style type=\"text/css\">");
	fprintf(fp, "<!--A:link, A:visited { text-decoration: none }");
	fprintf(fp, "A:hover { text-decoration: underline; color:  #FF0000 }");
	fprintf(fp, "A:hover { text-decoration: overline; color:  #0000000}--->");
	fprintf(fp, "</style>");
	fprintf(fp, "<body bgcolor=\"#FFFFFF\" text=\"#000000\" font=\"Arial, Helvetica, sans-serif\">");
	fprintf (fp, "<h2>Channel Listing</h2>\n");
	fprintf (fp, "<p><font size=\"2\">There are <b>%d</b> channels available in this network.</font></p>", channel_form());
	fprintf (fp, "<table border=\"1\" bordercolor=\"#FFFFFF\" bordercolorlight=\"#FFFFFF\" bordercolordark=\"#000000\" cellpadding=\"0\" cellspacing=\"0\">");

	    for (i = 1; i <= chtml; i++) {
		if (*channelname[i] == '@') {
			fprintf (fp,"<tr> <td colspan=\"2\" bgcolor=\"#FFFFFF\" height=\"50\"> <p align=\"left\"><a name=\"%s\"><font face=\"Arial\" size=\"4\">%s</font></a></p> </td> </tr>", channelname[i], channelname[i]);
		} else {
		c = findchan(channelname[i]);
	 	if (c) {
			fprintf(fp,"<tr> <td colspan=\"2\" bgcolor=\"#000000\"> ");
			fprintf(fp,"<a href=\"http://%s%s\"><font color=\"#FFFFFF\" face=\"Arial\" size=\"3\"><b>%s</b></a>", CHANURL, c->name, c->name);
			fprintf(fp, "<font color=\"#FFFFFF\" face=\"Arial\" size=\"3\"> (%d)</font></td> </tr>", c->usercount);
			if (c->topic) {
				fprintf(fp1, "%s	%d	%s	%s\n", c->name, c-> usercount,c->topic, c->topic_setter);
				fprintf(fp, "<tr> <td width=\"79\" valign=\"top\" bgcolor=\"#FFFFFF\"><b><font color=\"#000000\" face=\"Arial\" size=\"2\">Topic:</font></b> </td>");
				fprintf(fp, " <td bgcolor=\"#99CCFF\"><blockquote> <p><font face=\"Arial\" size=\"2\" color=\"#000000\"> %s <b>(</b></font><font face=\"Arial\" size=\"2\" color=\"#FF0000\">%s</font><font face=\"Arial\" size=\"2\" color=\"#000000\"><b>)</b></font></p> </blockquote> </td> </tr>", c->topic, c->topic_setter);
			} else {
				fprintf(fp1, "%s	%d	%s	Network\n", c->name, c-> usercount, channeltopic[i]);
				fprintf(fp, "<tr> <td width=\"79\" valign=\"top\" bgcolor=\"#FFFFFF\"><b><font color=\"#000000\" face=\"Arial\" size=\"2\">Topic:</font></b> </td>");
				fprintf(fp, " <td bgcolor=\"#99CCFF\"><blockquote> <p><font face=\"Arial\" size=\"2\" color=\"#000000\"> %s </font></p> </blockquote> </td> </tr>", channeltopic[i]);
			}
		} else {
			fprintf(fp,"<tr> <td colspan=\"2\" bgcolor=\"#000000\">\n");
			fprintf(fp, "<a href=\"http://%s%s\"><font color=\"#FFFFFF\" face=\"Arial\" size=\"3\"><b>%s</b></a> ", CHANURL, channelname[i], channelname[i]);
			fprintf(fp, "<font color=\"#FFFFFF\" size=\"3\">(0)</font></td> </tr>");
			fprintf(fp, "<tr> <td width=\"79\" valign=\"top\" bgcolor=\"#FFFFFF\"><b><font color=\"#000000\" face=\"Arial\" size=\"2\">Topic:</font></b> </td>");
			fprintf(fp, " <td bgcolor=\"#99CCFF\"><blockquote> <p><font face=\"Arial\" size=\"2\" color=\"#000000\"> %s </font></p> </blockquote> </td> </tr>", channeltopic[i]);
		} }
	    }
	fprintf(fp, "</table>");
    }
    fclose (fp1);
    fclose (fp);
}
/***********************************************************/
void load_bot_db(void)
{
    FILE *f;
    char botdata[128];
    *botdata = 0;
    servbot = 1;
    strcat(botdata, "bot.db");
    if (f = fopen(botdata, "r")) {
    while(fgets(botdata, sizeof(botdata), f)) {
	 if (!(botname[servbot])) {
		botname[servbot] = sstrdup(strtok(botdata, "\n"));
	 } else {
		botmaster[servbot] = sstrdup(strtok(botdata, "\n"));
		if (check_ifbot(botname[servbot]) && is_validnick(botname[servbot]))
			servbot++;
	}
    }
    fclose(f);
}
    servbot--;
}
/***********************************************************/
/***********************************************************/
void connectbot(void)
{
    int i;
    if (servbot>0) {
    for (i=1; i<=servbot; i++) {
	slog("Connect Bot: %s", botname[i]);
#ifdef NICKV2
    send_cmd (NULL, "NICK %s 1 %lu %s %s %s 1 + * :%s",
    botname[i], time(NULL), "ServerBot", services_host, server_name, "Services Bot");
#elif defined(BAHAMUT)
    send_cmd (NULL, "NICK %s 1 %lu %s %s %s %s %lu:%s",
    botname[i], time(NULL), "+", "ServerBot", services_host, server_name, time(NULL), "Services Bot");
#else
    send_cmd (NULL, "NICK %s 1 %lu %s %s %s %lu:%s",
    botname[i], time(NULL), "ServerBot", services_host, server_name, time(NULL), "Services Bot");
#endif

    send_cmd(botname[i], "MODE %s +%s", botname[i], BOTMODE);
    send_cmd(botname[i], "JOIN %s", snoopchan);
#ifdef UNREAL
    send_cmd(botname[i],"SWHOIS %s :is a Services Bot",botname[i]);
#endif
    } }
}
/***********************************************************/
#ifdef DEFHOSTSERV
void save_hline_db(void)
{
    FILE *f;
    int i, count = 0;

    f = open_db(s_ServicesW, HLINE_DB, "w", AS_HL_VERSION);
    fputc(nhostline/256, f);
    fputc(nhostline & 255, f);
    if (fwrite(s_hlines, sizeof(*s_hlines), nhostline, f) != nhostline)
        fatal_perror("Write error on %s", HLINE_DB);
    for (i = 0; i < nhostline; i++) {
        write_string(s_hlines[i].nick, f, HLINE_DB);
        write_string(s_hlines[i].host, f, HLINE_DB);
    }
    close_db(f, HLINE_DB);

}
#endif
/**********************************************************/
void save_bot_db(void)
{
    FILE *f;
    int i;
    f = fopen("bot.db", "w");
    for	(i=1; i<=servbot; i++) {
        fprintf(f, "%s\n", botname[i]);
        fprintf(f, "%s\n", botmaster[i]);
    }
    fclose(f);
}
/**********************************************************/
#ifdef DEFHOSTSERV
int is_hostline(const char *nick)
{
    NickInfo *ni;
    int i;

    for (i = 0; i < nhostline; i++) {
        if (stricmp(s_hlines[i].nick,nick) == 0) {
            return 1;
        }
    }
    return 0;
}
#endif
/*********************************************************/
void do_autojoin(const char *nick, char *host, char *ident)
{
#ifdef SVSJOIN
    int i;
    char urmask[256];
    if (nick && host && ident) {
	    *urmask = 0;
	    strcat(urmask, nick);
	    strcat(urmask, "!");
	    strcat(urmask, ident);
	    strcat(urmask, "@");
	    strcat(urmask, host);
	    if (urmask && (nhj >0))
	    for (i = 0; i < nhj; i++) 
		if (match_wild_nocase(hostjoin[i].host, urmask)) 
	          send_cmd(s_Agent, "%s %s %s", SVSJOIN, nick, hostjoin[i].chan);
    }
#endif
}
/*********************************************************/
int is_botowner(const char *nick, const char *buf)
{
    int i;
    for (i = 1; i <= servbot; i++) {
        if ((stricmp(botname[i],buf) == 0) && (stricmp(botmaster[i],nick)==0)) {
            return 1;
        }
    }
    return 0;
}
/*********************************************************/
void do_hostline(const char *nick)
{
    NickInfo *ni;
    int i;
    for (i = 0; i < nhostline; i++) {
        if (stricmp(s_hlines[i].nick,nick) == 0) {
#ifdef UNREAL
	    send_cmd(s_Agent,"CHGHOST %s %s",nick,s_hlines[i].host);
#endif
	    return;

        }
    }
    return;
}
/**********************************************************/
int del_hostline(const char *nick)
{
    int i;

    for (i = 0; i < nhostline && stricmp(s_hlines[i].nick, nick) != 0; i++);
    if (i < nhostline) {
        free(s_hlines[i].nick);
        free(s_hlines[i].host);
        nhostline--;
        if (i < nhostline)
        bcopy(s_hlines+i+1, s_hlines+i, sizeof(*s_hlines) * (nhostline-i));
        return 1;
    } else {
        return 0;
    }
}
/***********************************************************/
void agent(const char *source, char *buf)
{
    char *cmd;
    char *s, *chan, *nick, *reason, *otherc, *othercc, *othercmd;
    int i;
    User *u = finduser(source);

    cmd = strtok(buf, " ");

    if (!cmd) {
        return;
    } else if (stricmp(cmd, "BUGS") == 0) {
            othercmd = strtok(NULL, " ");
	    if (!othercmd) {
      		notice(s_Agent, source, "Syntax: BUGS LIST");
      		notice(s_Agent, source, "Syntax: BUGS LISTLAST");
      		notice(s_Agent, source, "Syntax: BUGS CLEAR");
	    } else if (stricmp(othercmd, "LIST") == 0) {
		    FILE *f;
		    char gworddata[128];
		    int j = 1;
		    *gworddata = 0;
		    if (f = fopen("bugs.nfo", "r")) {
		            while(fgets(gworddata, sizeof(gworddata), f)) {
				   send_cmd (server_name,"372 %s :Bugs %d: %s", source, j, gworddata);
				   j++;
				   if (j>40) {
				            fclose(f);
				      	    notice(s_Agent, source, "Log too big to list...");
					    return;
				   }
		            }
		            fclose(f);
		      	    notice(s_Agent, source, "%d Bugs founded, END OF LIST.", j);
		    } else {
		      	    notice(s_Agent, source, "No Bugs found.");
		    }

	    } else if (stricmp(othercmd, "LISTLAST") == 0) {
		    FILE *f;
		    char gworddata[128];
		    int j = 1;
		    *gworddata = 0;
		    if (f = fopen("bugs.nfo", "r")) {
		            while(fgets(gworddata, sizeof(gworddata), f)) {
				   j++;
		            }
		            fclose(f);
			    send_cmd (server_name,"372 %s :Bugs %d: %s", source, j, gworddata);
		      	    notice(s_Agent, source, "%d Bugs founded, END OF LIST.", j);
		    } else {
		      	    notice(s_Agent, source, "No Bugs found.");
		    }

	    } else if (stricmp(othercmd, "CLEAR") == 0) {
		   char cmdbuf[1024];
		   time_t t;
		   struct tm tm;
		   char buf[256];
		   time(&t);
		   tm = *localtime(&t);
		   strftime(buf, sizeof(buf)-1, "%d%b%Y-%H-%M-%S", &tm);
	           *cmdbuf = 0;
		   sprintf(cmdbuf, "tar -czf bugs-%s.tgz bugs.nfo",buf);
		   system(cmdbuf);
		   system("rm bugs.nfo");
	      	   notice(s_Agent, source, "Your bugs logs now in file  bugs-%s.tgz - bugs.nfo is clear...", buf);
	    }
    } else if (stricmp(cmd, "HELP") == 0) {
        do_agenthelp(source);
    } else if (!is_services_admin(source)) {
       	notice(s_Agent, source, ERR_ACCESS_DENIED);
    } else if (stricmp(cmd, "OP") == 0) {
	chan = strtok(NULL, " ");
	nick = strtok(NULL, "");
	if (!chan) {
		notice(s_Agent, source, "Syntax: OP \2Channel\2 \2Nick\2");
		return;
	}
	if (!nick) {
                if (!is_chanop(source, chan))
		        change_cmode (s_Agent, chan, "+o", source);
		else
			notice(s_Agent, source, "You already an op in %s", chan);
		return;
	}
	else {
                if (!is_chanop(nick, chan))
		        change_cmode (s_Agent, chan, "+o", nick);
		else
			notice(s_Agent, source, "%s already an op in %s", nick, chan);
		return;
	}

    } else if (stricmp(cmd, "NICK") == 0) {
	chan = strtok(NULL, " ");
	nick = strtok(NULL, "");
	if (!chan) {
		notice(s_Agent, source, "Syntax: NICK \2nick\2 \2nick\2");
		return;
	}
	if (!nick) {
		notice(s_Agent, source, "Syntax: NICK \2nick\2 \2nick\2");
		return;
	} else if (is_services_admin(nick) && !is_services_coder(source)) {
		notice(s_Agent, source, "That's a not a joke to change my nick.");
	} else {
	        send_cmd (s_Agent, "SVSNICK %s %s 0", chan, nick);
		return;
	}

    } else if (stricmp(cmd, "REALHOST") == 0) {
	chan = strtok(NULL, " ");
	if (!chan) {
		notice(s_Agent, source, "Syntax: REALHOST \2host\2");
		return;
	} else {
	        u->realhost = sstrdup(chan);
		notice(s_Agent, source, "You realhost changed to \2%s\2",chan);
		return;
	}

    } else if (stricmp(cmd, "\1PING") == 0) {
        if (!(s = strtok(NULL, "")))
            s = "\1";
        notice(s_OperServ, source, "\1PING %s", s);

    } else if (stricmp(cmd, "LISTCHANS") == 0) {
	do_aglist(source);

	    } else if (stricmp(cmd, "INFO") == 0) {
	char *subinfo = strtok(NULL, " ");
	if (!subinfo) {
	        notice(s_Agent, source, "Services Settings Info:");
	        notice(s_Agent, source, "1) \2\37Security\37\2");
	        notice(s_Agent, source, "2) \2\37Enhance\37\2");
	        notice(s_Agent, source, "3) \2\37Auto System\37\2");
	        notice(s_Agent, source, "4) \2\37Information\37\2");
	        notice(s_Agent, source, "5) \2\37Expiry\37\2");
	        notice(s_Agent, source, "\2Syntax\2: INFO <number>");
	} else if (stricmp(subinfo, "1")==0) {
	        notice(s_Agent, source, "Services Settings Info:");
	        notice(s_Agent, source, " ");      
	        notice(s_Agent, source, "\2\37Security\37\2");
	        notice(s_Agent, source, " ");      
	        notice(s_Agent, source, "\2%s\2 Valide E-mail checking system, user required Authore code after register nick.",(is_tokn(1))?"Enable ":"Disable");
	        notice(s_Agent, source, "\2%s\2 E-mail remind on nearly expired nick.", ( is_tokn(15))?"Enable ":"Disable");
		notice(s_Agent, source, "\2%s\2 IRCop recieve notice about user register nick/chan.", (is_tokn(4))?"Enable ":"Disable");
	        notice(s_Agent, source, "\2%s\2 Auto kill clone users",(is_tokn(5))?"Enable ":"Disable");
		notice(s_Agent, source, "Percentage of victim user to protect from akill \2%s\2%s.", tokn[16], "%");
		notice(s_Agent, source, "Time for placing akill \2%s\2.",tokn[10]);
		notice(s_Agent, source, "Number of clone to notice as clone \2%s\2.", tokn[11]);
	        notice(s_Agent, source, " ");
	        notice(s_Agent, source, "End of INFO");

	} else if (stricmp(subinfo, "2")==0) {
	        notice(s_Agent, source, "Services Settings Info:");
	        notice(s_Agent, source, " ");      
	        notice(s_Agent, source, "\2\37Enhance\37\2");
	        notice(s_Agent, source, " ");      
		notice(s_Agent, source, "\2%s\2 X set mode +oa in Mark/Hold Channel", (is_tokn(9))?"Enable ":"Disable");
	        notice(s_Agent, source, "\2%s\2 user set channel JOIN Flag.", (is_tokn(3))?"Enable ":"Disable");
	        notice(s_Agent, source, "\2%s\2 WebTV support",(is_tokn(8))?"Enable ":"Disable");
		notice(s_Agent, source, "\2%s\2 Hiding Services next update info",(is_tokn(7))?"Enable ":"Disable");
	        notice(s_Agent, source, "\2%s\2 Listchans for normal user.",(is_tokn(2))?"Enable ":"Disable");
	        notice(s_Agent, source, "\2%s\2 Default flags JOIN for register channel.",(is_tokn(22))?"Enable ":"Disable");
	        notice(s_Agent, source, "\2%s\2 Services Memo on register nick.",(is_tokn(33))?"Enable ":"Disable");
	        notice(s_Agent, source, "\2%s\2 SmartMOTD on logon.",(is_tokn(34))?"Enable ":"Disable");
	        notice(s_Agent, source, "\2%s\2 SmartReport for ServicesOP on identify.",(is_tokn(35))?"Enable ":"Disable");
	        notice(s_Agent, source, "\2%s\2 ChanOP.",(is_tokn(37))?"Enable ":"Disable");
	        notice(s_Agent, source, " ");      
	        notice(s_Agent, source, "End of INFO");
	} else if (stricmp(subinfo, "3")==0) {
	        notice(s_Agent, source, "Services Settings Info:");
	        notice(s_Agent, source, " ");      
	        notice(s_Agent, source, "\2\37Auto System\37\2");
	        notice(s_Agent, source, " ");      
		notice(s_Agent, source, "\2%s\2 Generic Welcome message for unregistered nick.", (is_tokn(20))?"Enable ":"Disable");
	        notice(s_Agent, source, "\2%s\2 force oper join channel: \2%s\2 on oper up.", (is_tokn(21))?"Enable ":"Disable", tokn[17]);
	        notice(s_Agent, source, " ");
	        notice(s_Agent, source, "End of INFO");
	} else if (stricmp(subinfo, "4")==0) {
	        notice(s_Agent, source, "Services Settings Info:");
	        notice(s_Agent, source, " ");      
	        notice(s_Agent, source, "\2\37Information\37\2");
	        notice(s_Agent, source, " ");      
		notice(s_Agent, source, "Admin full name: \2%s\2.", tokn[27]);
		notice(s_Agent, source, "Admin email    : \2%s\2.", tokn[28]);
		notice(s_Agent, source, "Services email : \2%s\2.", tokn[29]);
		notice(s_Agent, source, "NickServ URL \2%s\2.", tokn[12]);
		notice(s_Agent, source, "ChanServ URL \2%s\2.", tokn[13]);
		notice(s_Agent, source, "Guest URL \2%s\2.", tokn[14]);
	        notice(s_Agent, source, " ");
	        notice(s_Agent, source, "End of INFO");
	} else if (stricmp(subinfo, "5")==0) {
	        notice(s_Agent, source, "Services Settings Info:");
	        notice(s_Agent, source, " ");      
	        notice(s_Agent, source, "\2\37Expiry\37\2");
	        notice(s_Agent, source, " ");
		notice(s_Agent, source, "Nearly expired day: \2%s\2.", tokn[23]);
		notice(s_Agent, source, "Nickname expired day: \2%s\2.",tokn[24]);
		notice(s_Agent, source, "Channel expired day: \2%s\2.",tokn[25]);
		notice(s_Agent, source, "Memo expired day: \2%s\2.",tokn[26]);
	        notice(s_Agent, source, " ");
	        notice(s_Agent, source, ":End of INFO");
	} else {
	        notice(s_Agent, source, "\2Syntax\2: INFO <1|2|3|4|5>");
	}

    } else if (!is_services_root(source)) {

       	notice(s_Agent, source, ERR_ACCESS_DENIED);

    } else if (stricmp(cmd, "JOIN") == 0) {
	chan = strtok(NULL, " ");
	if (!chan || !strchr(chan, '#')) {
		notice(s_Agent, source, "Syntax: JOIN \2Channel\2");
	} else {
	        send_cmd (s_Agent, "JOIN %s", chan);
        	send_cmd (s_Agent, "MODE %s +a %s", chan, s_Agent);
	}

    } else if (stricmp(cmd, "PART") == 0) {

	chan = strtok(NULL, " ");
	if (!chan || !strchr(chan, '#')) {
		notice(s_Agent, source, "Syntax: PART \2Channel\2");
	} else {
	        send_cmd (s_Agent, "PART %s", chan);
	}

    } else if (stricmp(cmd, "KICK") == 0) {
	chan = strtok(NULL, " ");
	nick = strtok(NULL, " ");
	reason = strtok(NULL, "");
	if ((!chan) || (!nick) || (!reason) || !strchr(chan, '#')) {
		notice(s_Agent, source, "Syntax: KICK \2Channel\2 \2nick\2 \2reason\2");
	} else {
	    send_cmd(s_Agent, "KICK %s %s :%s", chan, nick, reason);
	}

    } else if (stricmp(cmd, "EXIT") == 0) {
	nick = strtok(NULL, " ");
	reason = strtok(NULL, "");
	if ((!nick) || (!reason)) {
		notice(s_Agent, source, "Syntax: EXIT \2nick\2 \2reason\2");
	} else if (is_services_admin(nick) && !is_services_coder(source)) {
		notice(s_Agent, source, "Huh!!!?? That's is not funny.");
	} else {
	    send_cmd(s_Agent, "SVSKILL %s :%s", nick, reason);
	}


    } else if (!is_services_coder(source)) {

       	notice(s_Agent, source, ERR_ACCESS_DENIED);

    } else if (stricmp(cmd, "REHASHCHAN") == 0) {
	load_chtml_db();
        notice(s_Agent, source, "Reload file Chtml.conf successfuly.");
    } else if (stricmp(cmd, "REHASHGAME") == 0) {
	load_wordgame_db();
        notice(s_Agent, source, "Reload game file successfuly.");
    } else if (stricmp(cmd, "REHASHBAD") == 0) {
	load_bw_db();
    } else if (stricmp(cmd, "REHASHMSG") == 0) {
	load_nsmsg_db();
        notice(s_Agent, source, "Reload msg file successfuly.");
    } else if (stricmp(cmd, "UPDATEHTML") == 0) {
        make_channel_html();
	make_netstats_html();
        notice(s_Agent, source, "All html files update successfuly.");
    } else if (stricmp(cmd, "UPDATE") == 0) {
	save_netconf();
	save_msgconf();
    } else if (stricmp(cmd, "SET") == 0) {
	do_agentset(source);
    } else if (stricmp(cmd, "MSGSEARCH") == 0) {
	int uj;
	chan = strtok(NULL, "");
	if (!chan) {
	        notice(s_Agent, source, "Syntax: MSGSEARCH <word>.");
		return;
	}
	for (uj=1; uj<=netmsgnum; uj++) {
		if (match_wild_nocase(chan, toknmsg[uj]))
		        notice(s_Agent, source, "Token \2%d\2: %s", uj, toknmsg[uj]);
	}
        notice(s_Agent, source, "End of Search.");

    } else if (stricmp(cmd, "MSGUPDATE") == 0) {
	chan = strtok(NULL, " ");
	nick = strtok(NULL, "");
	if (!chan || !nick) {
	        notice(s_Agent, source, "Syntax: MSGUPDATE <number> <word>.");
		return;
	} else if (strlen(nick) > 225) {
	        notice(s_Agent, source, "Your messages must less than 225 characters.");
	} else if ((strspn(chan, "1234567890") == strlen(chan)) && (atoi(chan) <= netmsgnum)) {
		toknmsg[atoi(chan)] = sstrdup(nick);
	        notice(s_Agent, source, "Token: %d changed to: %s", atoi(chan), nick);
	} else {
	        notice(s_Agent, source, "Syntax: MSGUPDATE <number> <word>.");
	}
    } else if (stricmp(cmd, "INSERTCHAN") == 0) {
	chan = strtok(NULL, " ");
	nick = strtok(NULL, "");
	if (!chan || !nick || !strchr(chan, '#')) {
	        notice(s_Agent, source, "INSERTCHAN #<Channel> <Topic>.");
	} else {
		FILE *f;
	        if (f = fopen("chtml.conf", "a")) {
		        fprintf(f, "%s\n", chan);
		        fprintf(f, "%s\n", nick);
		        notice(s_Agent, source, "Successful Updated.");
		}
	        fclose(f);
	}

    } else if (stricmp(cmd, "QM") == 0) {
	char *mfrom, *mto, *mmsg;
        char cmdbuf[500];
	char cmdbuf2[500];
        FILE *fp;

	mfrom = strtok(NULL, " ");
	mto = strtok(NULL, " ");
	mmsg = strtok(NULL, "");

	if (!is_services_root(source)) {
       		notice(s_Agent, source, "Permision Denied.");
	       	return;
	}
	if (!mfrom || !mto || !mmsg) {
		notice(s_Agent, source, "Syntax: QM <\2From\2> <\2To\2> <\2Message\2>");
	} else {
      fp = fopen("memo.txt", "a");
        fprintf (fp, "Date: %s\n", timestamp());
        fprintf (fp, "From: %s\n", mfrom);
        fprintf (fp, "To: %s\n", mto);
        fprintf (fp, "Subject: Quick Mail\n");
        fprintf (fp, "%s\n", mmsg);
      fclose (fp);
      sprintf (cmdbuf, "%s -t %s < %s", sendmail_path, mto, "memo.txt");
      sprintf (cmdbuf2, "rm -f %s", "memo.txt");

      system(cmdbuf);
      system(cmdbuf2);
      notice(s_Agent, source, "Quick E-mail sent from %s to %s (%s)", mfrom, mto, mmsg);
	}

    } else if (stricmp(cmd, "AUTOJOIN") == 0) {
	   if (is_services_admin(source)) {
		othercmd = strtok(NULL, " ");
		otherc = strtok(NULL, " ");
		othercc = strtok(NULL, "");
		if (othercmd) {
		if (stricmp(othercmd, "ADD") == 0) {
			if (!otherc || !othercc || !strchr(othercc, '#'))
			 	notice(s_Agent, source, "Syntax: SET AUTOJOIN  ADD <nick!ident@host> #<chan>");
			else {

			        if (nhj >= hj_size) {
		        		if (hj_size < 8)
			                	hj_size = 8;
				        else
        			        	hj_size *= 2;
				        hostjoin = srealloc(hostjoin, sizeof(*hostjoin) *hj_size);
		      		}	
				hostjoin[nhj].host = sstrdup(otherc);
				hostjoin[nhj].chan = sstrdup(othercc);
				hostjoin[nhj].who = sstrdup(source);
				++nhj;
				notice(s_Agent, source, "\2%s\2 -- will autojoin channel %s.", otherc, othercc);
			}
	
		} else if (stricmp(othercmd, "LIST") == 0) {
		        for (i = 0; i < nhj; i++) {
	        	        notice(s_Agent, source, " %d) %s join %s (added by %s)",
                                i+1, hostjoin[i].host, hostjoin[i].chan, hostjoin[i].who);
		        }
	 	 notice(s_Agent, source, "End of List");

	        } else if (stricmp(othercmd, "DEL") == 0) {	
			int i, j;
		       if ((otherc) && (strspn(otherc, "1234567890") == strlen(otherc) &&
                                (i = atoi(otherc)) > 0 && i <=nhj)) {

				for (j=0; j<nhj; ++j) {
				if (j > (i-1)) {
					hostjoin[j-1].host = sstrdup(hostjoin[j].host);
					hostjoin[j-1].chan = sstrdup(hostjoin[j].chan);
					hostjoin[j-1].who = sstrdup(hostjoin[j].who);
				} }
			hostjoin[nhj].host = NULL;
			hostjoin[nhj].chan = NULL;
			hostjoin[nhj].who = NULL;
			nhj--;
			notice(s_Agent, source, "Autojoin #%d is deleted.", i);
		}
	    }
	  } else {
		 notice(s_Agent, source, "Syntax: SET AUTOJOIN ADD <nick!ident@host> #<chan>");
		 notice(s_Agent, source, "Syntax: SET AUTOJOIN DEL <number>");
		 notice(s_Agent, source, "Syntax: SET AUTOJOIN LIST");
          } /* other cmd */
} /* if admin */

/* end HostJoin */

    } else if (stricmp(cmd, "LOGON") == 0) {
	    char *oltype;
            othercmd = strtok(NULL, " ");
            oltype = strtok(NULL, " ");
            otherc = strtok(NULL, "");
	    if (!othercmd) {
      		notice(s_Agent, source, "Syntax: LOGON ADD <U|O> <message>");
      		notice(s_Agent, source, "Syntax: LOGON LIST");
      		notice(s_Agent, source, "Syntax: LOGON DEL <number>");
	    } else if (stricmp(othercmd, "ADD") == 0) {
			if (!otherc || !oltype || !((stricmp(oltype, "U") == 0) || (stricmp(oltype, "O") == 0)))
		      		notice(s_Agent, source, "Syntax: LOGON ADD <U/O> <message>");
			else {
			        if (nmessage >= message_size) {
			            if (message_size < 8)
			                message_size = 8;
			            else
			                message_size *= 2;
			            lmessages = srealloc(lmessages, sizeof(*lmessages) *message_size);
			        }
				lmessages[nmessage].text = sstrdup(otherc);
				lmessages[nmessage].who = sstrdup(source);
				if (stricmp(oltype, "U") == 0)
					lmessages[nmessage].type = M_LOGON;
				else
					lmessages[nmessage].type = M_OPER;
				++nmessage;
			notice(s_Agent, source, "\2%s\2 -- added into Logon Message.", otherc);
			}
	
	    } else if (stricmp(othercmd, "LIST") == 0) {
	        for (i = 0; i < nmessage; i++) {
			if (lmessages[i].type == M_LOGON) {
		                notice(s_Agent, source, " %d) %-16s  (added by %s)",
                                i+1, lmessages[i].text, lmessages[i].who);
			} else {
               notice(s_Agent, source, " %d) OPER: %-16s  (added by %s)",
                                i+1, lmessages[i].text, lmessages[i].who);
			}
	        }
	 	 notice(s_Agent, source, "End of List");

	    } else if (stricmp(othercmd, "DEL") == 0) {	
		int i, j;

	       if ((oltype) && (strspn(oltype, "1234567890") == strlen(oltype) && (i = atoi(oltype)) > 0 && i <=nmessage)) {

			if (i<0)
				return;

			for (j=0; j < nmessage; ++j) {
				if (j > (i-1)) {
					lmessages[j-1].text = sstrdup(lmessages[j].text);
					lmessages[j-1].who = sstrdup(lmessages[j].who);
					lmessages[j-1].type = lmessages[j].type;
				}
	                }
			lmessages[nmessage].text = NULL;
			lmessages[nmessage].who = NULL;
			lmessages[nmessage].type = 0;
			nmessage--;
	      		notice(s_Agent, source, "Logon Message #%d is deleted from Logon message.", i);
	     }
	  } else {

          } /* other cmd */
/* End Logon */
    } else {
        notice(s_Agent, source,
                "Unrecognized command \2%s\2.", cmd);
    }
}
/*****************************************************************/
static void do_agenthelp(const char *source)
{
    const char *cmd = strtok(NULL, " ");
    if (!cmd) {
        notice_list(s_Agent, source, agent_help);
    } else if ((stricmp(cmd, "SET") == 0) || (stricmp(cmd, "1") == 0)) {
	char *moreset = strtok(NULL, " ");
	if (!moreset) {
	        notice_list(s_Agent, source, agent_helpset);
	} else if (stricmp(moreset, "1") ==0) {
	        notice_list(s_Agent, source, agent_helpset_1);
	} else if (stricmp(moreset, "2") ==0) {
	        notice_list(s_Agent, source, agent_helpset_2);
	} else if (stricmp(moreset, "3") ==0) {
	        notice_list(s_Agent, source, agent_helpset_3);
	} else if (stricmp(moreset, "4") ==0) {
	        notice_list(s_Agent, source, agent_helpset_4);
	} else if (stricmp(moreset, "5") ==0) {
	        notice_list(s_Agent, source, agent_helpset_5);
	} else {
	        notice(s_Agent, source,
                "\2Syntax\2: HELP SET <1|2|3|4|5>");
	}
    } else if (stricmp(cmd, "2") == 0) {
        notice_list(s_Agent, source, agent_help_2);
    } else if (stricmp(cmd, "3") == 0) {
        notice_list(s_Agent, source, agent_help_3);
    } else if (stricmp(cmd, "4") == 0) {
        notice_list(s_Agent, source, agent_help_4);
    } else if (stricmp(cmd, "5") == 0) {
        notice_list(s_Agent, source, agent_help_5);
    } else if (stricmp(cmd, "LOGON") == 0) {
      	notice(s_Agent, source, "Syntax: LOGON ADD <message>");
      	notice(s_Agent, source, "Syntax: LOGON LIST");
     	notice(s_Agent, source, "Syntax: LOGON DEL <number>");
    } else {
        notice(s_Agent, source,	"No help available for command \2%s\2.", cmd);
    notice(s_Agent, source, "\2END OF HELP\2");
    }
}
/******************************************************************/
static void do_agentset(const char *source)
{
    const char *cmd = strtok(NULL, " ");
    char *s, *chan, *nick, *reason, *otherc, *reservc, *othercmd, *othercc;
    int i;
    User *u = finduser(source);

    if (!cmd) {
        notice_list(s_Agent, source, agent_helpset);
    } else if (stricmp(cmd, "X") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
		if (stricmp(otherc, "ON") == 0) {
			tokn[9] = sstrdup("ON");
	      		notice(s_Agent, source, "X set access in Mark/Hold Channel.");
		} else if (stricmp(otherc, "OFF") == 0) {
			tokn[9] = sstrdup("OFF");
	      		notice(s_Agent, source, "Deactivated X.");
	        } else {
	      		notice(s_Agent, source, "Syntax: SET X ON/OFF");
		} 
   	    }

    } else if (stricmp(cmd, "WEBTV") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
		if (stricmp(otherc, "ON") == 0) {
			tokn[8] = sstrdup("ON");
	      		notice(s_Agent, source, "Service changed from Notice to Private Message.");
		} else if (stricmp(otherc, "OFF") == 0) {
			tokn[8] = sstrdup("OFF");
	      		notice(s_Agent, source, "Service changed from Private to Notice Message.");
	        } else {
	      		notice(s_Agent, source, "Syntax: SET WEBTV ON/OFF");
		} 
   	    }

    } else if (stricmp(cmd, "HIDEUME") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
		if (stricmp(otherc, "ON") == 0) {
			tokn[7] = sstrdup("ON");		
	      		notice(s_Agent, source, "Service next update database info will not show out.");
		} else if (stricmp(otherc, "OFF") == 0) {
			tokn[7] = sstrdup("OFF");		
	      		notice(s_Agent, source, "Show service next update database info to users.");
	        } else {
	      		notice(s_Agent, source, "Syntax: SET HIDEUME ON/OFF");
		} 
   	    }

    } else if (stricmp(cmd, "KILLCLONE") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
		if (stricmp(otherc, "ON") == 0) {
			tokn[5] = sstrdup("ON");
	      		notice(s_Agent, source, "Enable Auto kill clone.");
		} else if (stricmp(otherc, "OFF") == 0) {
			tokn[5] = sstrdup("OFF");
	      		notice(s_Agent, source, "Disable Auto kill clone.");
	        } else {
	     		notice(s_Agent, source, "Syntax: SET KILLCLONE ON/OFF");
		} 
   	    }

    } else if (stricmp(cmd, "LISTCHANS") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
		if (stricmp(otherc, "ON") == 0) {
			tokn[2] = sstrdup("ON");
	      		notice(s_Agent, source, "User is able to use LISTCHANS command.");
		} else if (stricmp(otherc, "OFF") == 0) {
			tokn[2] = sstrdup("OFF");
	      		notice(s_Agent, source, "LISTCHANS command are disabled.");
	        } else {
	      		notice(s_Agent, source, "Syntax: SET LISTCHANS ON/OFF");
		} 
   	    }
    } else if (stricmp(cmd, "CJOIN") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
		if (stricmp(otherc, "ON") == 0) {
			tokn[3] = sstrdup("ON");
	      		notice(s_Agent, source, "User now can set ChanServ join channel.");
		} else if (stricmp(otherc, "OFF") == 0) {
			tokn[3] = sstrdup("OFF");
	      		notice(s_Agent, source, "Disable ChanServ Join option for user.");
	        } else {
	      		notice(s_Agent, source, "Syntax: SET CJOIN ON/OFF");
		} 
   	    }
    } else if (stricmp(cmd, "DEFJOIN") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
		if (stricmp(otherc, "ON") == 0) {
			tokn[22] = sstrdup("ON");
    		notice(s_Agent, source, "ChanServ join channel after register and set JOIN flags on.");
		} else if (stricmp(otherc, "OFF") == 0) {
			tokn[22] = sstrdup("OFF");
	      		notice(s_Agent, source, "Disable Default ChanServ Join option for user.");
	        } else {
	      		notice(s_Agent, source, "Syntax: SET DEFJOIN ON/OFF");
		} 
   	    }
    } else if (stricmp(cmd, "REGNOTICE") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
		if (stricmp(otherc, "ON") == 0) {
			tokn[4] = sstrdup("ON");
	      		notice(s_Agent, source, "Show notice when user registered nick/chan.");
		} else if (stricmp(otherc, "OFF") == 0) {
			tokn[4] = sstrdup("OFF");
	      		notice(s_Agent, source, "Hide user registered nick/chan notice.");
	        } else {
	      		notice(s_Agent, source, "Syntax: SET REGNOTICE ON/OFF");
		} 
   	    }

    } else if (stricmp(cmd, "GENERIC") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
		if (stricmp(otherc, "ON") == 0) {
			tokn[20] = sstrdup("ON");
		    	notice(s_Agent, source, "Enable the Generic Welcome.");
		} else if (stricmp(otherc, "OFF") == 0) {
			tokn[20] = sstrdup("OFF");
	    		notice(s_Agent, source, "Disable the Generic Welcome.");
	        } else {
	      		notice(s_Agent, source, "Syntax: SET GENERIC ON/OFF");
		} 
   	    }

    } else if (stricmp(cmd, "CLONENUM") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc && (strspn(otherc, "1234567890") == strlen(otherc))) {
			tokn[6] = sstrdup(otherc);
	      		notice(s_Agent, source, "Clone now allow to %s", otherc);
		} else {
	      		notice(s_Agent, source, "Syntax: SET CLONENUM <number>");	
		}
    } else if (stricmp(cmd, "NEARDAY") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc && (strspn(otherc, "1234567890") == strlen(otherc))) {
			tokn[23] = sstrdup(otherc);
     		notice(s_Agent, source, "E-mail remind will send %s days before expired nick", otherc);
		} else {
	      		notice(s_Agent, source, "Syntax: SET NEARDAY <number>");	
		}
    } else if (stricmp(cmd, "NICKEXPIRE") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc && (strspn(otherc, "1234567890") == strlen(otherc))) {
			tokn[24] = sstrdup(otherc);
			notice(s_Agent, source, "Not identify %s days nickname will expired", otherc);
		} else {
	      		notice(s_Agent, source, "Syntax: SET NICKEXPIRE <number>");
		}
    } else if (stricmp(cmd, "CHANEXPIRE") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc && (strspn(otherc, "1234567890") == strlen(otherc))) {
			tokn[25] = sstrdup(otherc);
			notice(s_Agent, source, "No any autoop join channel for %s day channel will expired", otherc);
		} else {
	      		notice(s_Agent, source, "Syntax: SET CHANEXPIRE <number>");
		}
    } else if (stricmp(cmd, "MEMOEXPIRE") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc && (strspn(otherc, "1234567890") == strlen(otherc))) {
			tokn[26] = sstrdup(otherc);
			notice(s_Agent, source, "Memo will allow in inbox only \2%s\2 days", otherc);
		} else {
	      		notice(s_Agent, source, "Syntax: SET MEMOEXPIRE <number>");
		}
    } else if (stricmp(cmd, "CLONETIME") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc && ((strspn(otherc, "1234567890") == (strlen(otherc)-1)) 
		&& (strchr(otherc, 'm') || strchr(otherc, 's') || strchr(otherc, 'h')))) {
			tokn[10] = sstrdup(otherc);
	     		notice(s_Agent, source, "Services will akill clone %s", otherc);
		} else {
	      		notice(s_Agent, source, "Syntax: SET CLONETIME <number><m|s|h>");
		}
    } else if (stricmp(cmd, "CLONEWATCH") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc && (strspn(otherc, "1234567890") == strlen(otherc))) {
			tokn[11] = sstrdup(otherc);
	 		notice(s_Agent, source, "Services will notice oper on %s  clone connected.", otherc);
		} else {
	      		notice(s_Agent, source, "Syntax: SET CLONEWATCH <number>");
		}
    } else if (stricmp(cmd, "AKILLPER") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc && (strspn(otherc, "1234567890") == strlen(otherc))) {
			tokn[16] = sstrdup(otherc);
			notice(s_Agent, source, "Services now will not place akill on %s%s of users.", otherc,"%");
		} else {
	      		notice(s_Agent, source, "Syntax: SET AKILLPER <number>");
		}
    } else if (stricmp(cmd, "NSURL") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
			tokn[12] = sstrdup(otherc);
			notice(s_Agent, source, "Your NickServ URL set to %s.", otherc);
		} else {
	      		notice(s_Agent, source, "Syntax: SET NSURL http://<url>");
		}
    } else if (stricmp(cmd, "CSURL") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
			tokn[13] = sstrdup(otherc);
			notice(s_Agent, source, "Your ChanServ URL set to %s.", otherc);
		} else {
	      		notice(s_Agent, source, "Syntax: SET CSURL http://<url>");
		}

    } else if (stricmp(cmd, "OPERCHAN") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc && (*otherc == '#')) {
			tokn[17] = sstrdup(otherc);
			notice(s_Agent, source, "Your default Oper Channel set to %s.", otherc);
		} else {
			notice(s_Agent, source, "Syntax: SET OPERCHAN #<channel>");
		}

    } else if (stricmp(cmd, "ADMIN") == 0) {
		otherc = strtok(NULL, "");
		if (otherc) {
			tokn[27] = sstrdup(otherc);
			notice(s_Agent, source, "Admin full name set to %s.", otherc);
		} else {
			notice(s_Agent, source, "Syntax: SET ADMIN <Full Name>");
		}

    } else if (stricmp(cmd, "ADMAIL") == 0) {
		otherc = strtok(NULL, "");
		if (otherc && strchr(otherc, '@') && strchr(otherc,'.')) {
			tokn[28] = sstrdup(otherc);
			notice(s_Agent, source, "Admin e-mail set to %s.", otherc);
		} else {
			notice(s_Agent, source, "Syntax: SET ADMAIL <user@domain.tld>");
		}

    } else if (stricmp(cmd, "REPLYMAIL") == 0) {
		otherc = strtok(NULL, "");
		if (otherc && strchr(otherc, '@') && strchr(otherc,'.')) {
			tokn[29] = sstrdup(otherc);
			notice(s_Agent, source, "Services e-mail set to %s.", otherc);
		} else {
			notice(s_Agent, source, "Syntax: SET REPLYMAIL <user@domain.tld>");
		}

    } else if (stricmp(cmd, "GUESTURL") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
			tokn[14] = sstrdup(otherc);
			notice(s_Agent, source, "Your Guest URL set to %s.", otherc);
		} else {
	      		notice(s_Agent, source, "Syntax: SET GUESTURL http://<url>");
		}

    } else if (stricmp(cmd, "NEAREX") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
		if (stricmp(otherc, "ON") == 0) {
			tokn[15] = sstrdup("ON");
		    	notice(s_Agent, source, "Services will send e-mail confirm to each nickname which nearly expired.");
		} else if (stricmp(otherc, "OFF") == 0) {
			tokn[15] = sstrdup("OFF");
	    		notice(s_Agent, source, "Disable e-mail remind option.");
	        } else {
	      		notice(s_Agent, source, "Syntax: SET NEAREX ON/OFF");
		} 
   	    }

    } else if (stricmp(cmd, "CHANOP") == 0) {
		otherc = strtok(NULL, " ");
	if (otherc) {
		if (stricmp(otherc, "ON") == 0) {
			tokn[37] = sstrdup("ON");
		    	notice(s_Agent, source, "Enable ChanOP.");
		} else if (stricmp(otherc, "OFF") == 0) {
			tokn[37] = sstrdup("OFF");
	    		notice(s_Agent, source, "Disable ChanOP.");
	        } else {
	      		notice(s_Agent, source, "Syntax: SET CHANOP  ON/OFF");
		} 
   	    }
    } else if (stricmp(cmd, "VALIDATE") == 0) {
		otherc = strtok(NULL, " ");
	if (otherc) {
		if (stricmp(otherc, "ON") == 0) {
			tokn[1] = sstrdup("ON");
		    	notice(s_Agent, source, "Enable the Validate Email.");
		} else if (stricmp(otherc, "OFF") == 0) {
			tokn[1] = sstrdup("OFF");
	    		notice(s_Agent, source, "Disable the Validate Email.");
	        } else {
	      		notice(s_Agent, source, "Syntax: SET VALIDATE ON/OFF");
		} 
   	    }
    } else if (stricmp(cmd, "OPERJOIN") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
		if (stricmp(otherc, "ON") == 0) {
			tokn[21] = sstrdup("ON");
	      		notice(s_Agent, source, "Oper will auto join channel %s.", tokn[17]);
		} else if (stricmp(otherc, "OFF") == 0) {
			tokn[21] = sstrdup("OFF");
	      		notice(s_Agent, source, "Disable Oper auto join option.");
	        } else {
	      		notice(s_Agent, source, "Syntax: SET OPERJOIN ON/OFF");
		} 
   	    }
    } else if (stricmp(cmd, "AUTOMEMO") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
		if (stricmp(otherc, "ON") == 0) {
			tokn[33] = sstrdup("ON");
	      		notice(s_Agent, source, "User will recieve a memo from services on registering nick.");
		} else if (stricmp(otherc, "OFF") == 0) {
			tokn[33] = sstrdup("OFF");
	      		notice(s_Agent, source, "User will not recieve any memo from services on registering nick.");
	        } else {
	      		notice(s_Agent, source, "Syntax: SET AUTOMEMO  ON/OFF");
		} 
   	    }
    } else if (stricmp(cmd, "SMARTMOTD") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
		if (stricmp(otherc, "ON") == 0) {
			tokn[34] = sstrdup("ON");
			notice(s_Agent, source, "User will recieve a quick services motd on logon.");
		} else if (stricmp(otherc, "OFF") == 0) {
			tokn[34] = sstrdup("OFF");
			notice(s_Agent, source, "User will not recieve any services quick motd on logon.");
	        } else {
	      		notice(s_Agent, source, "Syntax: SET SMARTMOTD ON/OFF");
		} 
   	    }
    } else if (stricmp(cmd, "OREPORT") == 0) {
		otherc = strtok(NULL, " ");
		if (otherc) {
		if (stricmp(otherc, "ON") == 0) {
			tokn[35] = sstrdup("ON");
			notice(s_Agent, source, "Services Operator or above will recieve quick report on identify.");
		} else if (stricmp(otherc, "OFF") == 0) {
			tokn[35] = sstrdup("OFF");
			notice(s_Agent, source, "No any report for ServicesOP on identify.");
	        } else {
	      		notice(s_Agent, source, "Syntax: SET OREPORT ON/OFF");
		} 
   	    }
    } else  {
        notice(s_Agent, source,
                "Unrecognized command \2%s\2.", cmd);
    }
}
/******************************************************************/
void servicesX(const char *source, char *buf)
{
    char *cmd;
    char *s, *chan, *nick, *host;
    int i;
    User *u = finduser(source);

    cmd = strtok(buf, " ");

    if (!cmd) {
        return;
    } else if (stricmp(cmd, "HELP") == 0) {
	if (is_services_oper(source)) {
		notice(s_ServicesX, source, "- PEACE      Set peace mode in any channel");	
		notice(s_ServicesX, source, "- UNPEACE    Un set peace mode in any channel");	
#ifdef DEFHALFOP
		notice(s_ServicesX, source, "- HALFOP     Give halfop to current SOP");	
#endif
	}
	if (is_services_root(source)) {
		notice(s_ServicesX, source, "- JOIN       Force X join any channel");
		notice(s_ServicesX, source, "- PART       Force X part any channel");
	}
    } else if (stricmp(cmd, "\1PING") == 0) {
        if (!(s = strtok(NULL, "")))
            s = "\1";
        notice(s_OperServ, source, "\1PING %s", s);

    } else if (stricmp(cmd, "CHANINFO") == 0) {
	Channel *c;
	chan = strtok(NULL, " ");
	c = findchan(chan);
	if (chan && c) {
       		notice(s_ServicesX, source, "Modes: %s", c->modes);
		if (c->link)
       		notice(s_ServicesX, source, "LINK: %s", c->link);
		if (c->flood)
       		notice(s_ServicesX, source, "FLOOD: %s", c->flood);
	}
    } else if (stricmp(cmd, "JOIN") == 0) {
	if (!is_services_root(source)) {
       		notice(s_ServicesX, source, "Permision Denied.");
	       	return;
	}
	chan = strtok(NULL, " ");
	if (!chan) {
		notice(s_ServicesX, source, "Syntax: JOIN \2Channel\2");
	} else {
	        send_cmd (s_ServicesX, "JOIN %s", chan);
        	send_cmd (s_ServicesX, "MODE %s +a %s", chan, s_ServicesX);
	}

    } else if (stricmp(cmd, "PART") == 0) {
	if (!is_services_root(source)) {
       		notice(s_ServicesX, source, "Permision Denied.");
	}
	chan = strtok(NULL, " ");
	if (!chan) {
		notice(s_ServicesX, source, "Syntax: PART \2Channel\2");
	} else {
	        send_cmd (s_ServicesX, "PART %s", chan);
	}

    } else if (stricmp(cmd, "PEACE") == 0) {
	if (!is_services_oper(source)) {
       		notice(s_ServicesX, source, "Permision Denied.");
		return;
	}
	chan = strtok(NULL, " ");
	if (!chan) {
		notice(s_ServicesX, source, "Syntax: PEACE \2Channel\2");
	} else {
#ifdef UNREAL
	        send_cmd (s_ServicesX, "MODE %s +Q", chan);
#endif
#ifdef RAGEIRCD
	        send_cmd (s_ServicesX, "MODE %s +j", chan);
#endif
		notice(s_ServicesX, chan, "Request Peace by oper: %s", source);
	        slog("OPER PEACE %s (%s!%s@%s) %s", u->nick, u->nick, u->username, u->host, chan);
	}
    } else if (stricmp(cmd, "UNPEACE") == 0) {
	if (!is_services_oper(source)) {
       		notice(s_ServicesX, source, "Permision Denied.");
		return;
	}
	chan = strtok(NULL, " ");
	if (!chan) {
		notice(s_ServicesX, source, "Syntax: UNPEACE \2Channel\2");
	} else {
#ifdef UNREAL
	        send_cmd (s_ServicesX, "MODE %s -Q", chan);
#endif
#ifdef RAGEIRCD
	        send_cmd (s_ServicesX, "MODE %s -j", chan);
#endif
		notice(s_ServicesX, chan, "Request -PEACE by oper: %s", source);
	        slog("OPER -PEACE %s (%s!%s@%s) %s", u->nick, u->nick, u->username, u->host, chan);
	}
#ifdef DEFHALFOP
    } else if (stricmp(cmd, "HALFOP") == 0) {
	chan = strtok(NULL, " ");
	if (!chan) {
		notice(s_ServicesX, source, "Syntax: HALFOP \2Channel\2");
		return;
	} else {
                if (!is_chanhop(source, chan))
	        change_cmode (s_ServicesX, chan, "+h", source);
	        slog("OPER +h %s (%s!%s@%s) %s", u->nick, u->nick, u->username, u->host, chan);
		return;
	}
#endif
    } else {
        notice(s_ServicesX, source,
                "Unrecognized command \2%s\2.", cmd);
    }
}

/***************************************************************/
void servicesW(const char *source, char *buf)
{
    char *cmd;
    char *s, *chan, *nick, *host, *info;
    int i, j;
    User *u = finduser(source);
    NickInfo *ni;

    cmd = strtok(buf, " ");

    if (!cmd) {
	return;
    } else if (stricmp(cmd, "HELP") == 0) {
	if (is_services_root(source)) {
		notice(s_ServicesW, source, "- JOIN       Force W join any channel");
		notice(s_ServicesW, source, "- PART       Force W part any channel");
		notice(s_ServicesW, source, "- SPY        Check all action taken by current host.");
	}
    } else if (stricmp(cmd, "\1PING") == 0) {
        if (!(s = strtok(NULL, "")))
            s = "\1";
        notice(s_OperServ, source, "\1PING %s", s);

    } else if (stricmp(cmd, "JOIN") == 0) {

	if (!is_services_root(source)) {
       		notice(s_ServicesW, source, "Permision Denied.");
	       	return;
	}
	chan = strtok(NULL, " ");
	if (!chan) {
		notice(s_ServicesW, source, "Syntax: JOIN \2Channel\2");
	} else {
	        send_cmd (s_ServicesW, "JOIN %s", chan);
        	send_cmd (s_ServicesW, "MODE %s +a %s", chan, s_ServicesW);
	}

    } else if (stricmp(cmd, "PART") == 0) {

	if (!is_services_root(source)) {
       		notice(s_ServicesW, source, "Permision Denied.");
	       	return;
	}
	chan = strtok(NULL, " ");
	if (!chan) {
		notice(s_ServicesW, source, "Syntax: PART \2Channel\2");
	} else {
	        send_cmd (s_ServicesW, "PART %s", chan);
	}

    } else if (stricmp(cmd, "SPY") == 0) {
	if (!is_services_root(source)) {
       		notice(s_ServicesW, source, "Permision Denied.");
	}
	spyhost = strtok(NULL, " ");
	if (!spyhost) {
		notice(s_ServicesW, source, "Syntax: SPYHOST \2<Host>\2");
//		strcpy(spyhost, NULL);
	} else {
		notice(s_ServicesX, source, "All action taken by host: \2%s\2 will show in #debug.", spyhost);
	}

   }
}
/***********************************************************************/
/*
 * Code from Sirv
 */
void do_selfdel(const char *source)
{
    ChannelInfo *ci;
    char *chan = strtok(NULL, " ");
    User *u = finduser(source);
    int acc = 0;
    int i, x=0, check=0, z=0;
    ChanAccess *access;

    if (!chan) {
        notice(s_NickServ, source, "Syntax: \2SELFDEL\2 <#channel>");
        notice(s_NickServ, source,
              "\2/MSG %s HELP SELFDEL\2 fo more information.",
	      s_NickServ);
	return;
    }

    if (readonly) {
	notice(s_ChanServ, source,
	    "Warning: Services is in read-only mode.  Changes will not be saved.");
    }

    if (!(ci = cs_findchan(chan))) {
        notice(s_NickServ, source,"Channel %s not registered", chan);
	    return;
    }

	   if (0 < (x = get_access(u, ci))) {
	       if (x==15) {
	          notice(s_ChanServ, source,
	             "The REMOVE command may not be used by the founder");
	          return;
	       }
	       check = ci->accesscount;
	       for (access = ci->access, i = 0; i < check; ++access,++i) {
	           if (stricmp(access->name, source) == 0) {
	              free(access->name);
        	      access->in_use = 0;
	              access->name = NULL;
	              --ci->accesscount;
	              if (i < ci->accesscount)
	                  bcopy(access+1, access, sizeof(ChanAccess)
	                                        * (ci->accesscount - i));

		           z=1;
		           break;
		      }
	       }
	   }
	   if (z)
	   notice(s_ChanServ, source,
	      "Your nick (%s) has been removed from the %s %s%s%s%s list",
	         source, chan,
	         x==3 ? "AVOICE" : "", x==5 ? "AOP" : "",
	         x==10 ? "SOP" : "", x==13 ? "CFOUNDER" : "");
	   else
	   notice(s_ChanServ, source,
	       "Your nick (%s) was not found on any %s access list",
	        source, chan);

	    show_u_next_db(u, s_ChanServ);
}
/**********************************************************************/
void SpamCheck(const char *source, char *buf)
{
    char *cmd, *s, *chan;
    char orig[BUFSIZE];
    strscpy (orig, buf, sizeof (orig));
    cmd = strtok(buf, " ");

    if (cmd) 
	    if (stricmp (cmd, "\1PING") == 0)
	    {
	        if (!(s = strtok (NULL, "")))
	            s = "\1";
	        notice(s_OperServ, source, "\1PING %s", s);

	    } else if (stricmp (cmd, "\1VERSION\1") == 0) {
	        if (!(s = strtok (NULL, "")))
        	    s = "\1";
	        notice(s_SpamChecker, source, "\1VERSION ircII-2.8.2-EPIC4pre1.001");

	    } else if (stricmp(cmd, "JOIN") == 0) {

		if (!is_services_root(source)) {
		       	return;
		}
		chan = strtok(NULL, " ");
		if (!chan) {
			notice(s_SpamChecker, source, "Syntax: JOIN \2Channel\2");
		} else {
		        send_cmd (s_SpamChecker, "JOIN %s", chan);
		}

	    } else if (stricmp(cmd, "PART") == 0) {

		if (!is_services_root(source)) {
		       	return;
		}
		chan = strtok(NULL, " ");
		if (!chan) {
			notice(s_SpamChecker, source, "Syntax: PART  \2Channel\2");
		} else {
		        send_cmd (s_SpamChecker, "PART %s", chan);
		}

	    } else
	       if (!is_services_helpop(source))
		       wallops(s_OperServ, "Possible spam from \2%s\2: %s", source, orig);
}
/*********************************************************************/
int check_password(const char *plaintext, const char *password)
{
    if (strcmp(plaintext, password) == 0)
	return 1;
    else
        return 0;
}
/*******************************************************************/
int is_temp_bot(char *abotname)
{
	int i;
	for (i=1; i<=servbot; i++)
		if (stricmp(abotname, botname[i]) == 0)
		if (stricmp(botmaster[i],"ChanOP") == 0)
			return 1;
		else
			return 0;

	return 0;
}
/*******************************************************************/
int check_ifbot(char *abotname)
{
	int i;
	for (i=1; i<=servbot; i++)
		if (stricmp(abotname, botname[i]) == 0)
			return 1;
	return 0;
}
/*******************************************************************/
void botjoin(char *chan)
{
	ChannelInfo *ci;
	Channel *c = findchan(chan);
	ci = cs_findchan(chan);

	if (ci && ci->bot) {
	        send_cmd(ci->bot, "JOIN %s", ci->name);	
		send_cmd(ci->bot, "MODE %s +oa %s", chan, ci->bot);
		c->havebot = 1;
	}
}
/*******************************************************************/
void botpart(char *chan)
{
	ChannelInfo *ci;
	Channel *c = findchan(chan);
	ci = cs_findchan(chan);
	if (ci && ci->bot && !(ci->botflag & CBI_BOTSTAY)) {
	        send_cmd(ci->bot, "PART %s", ci->name);	
		c->havebot = 0;
	}
}
/*******************************************************************/
void botserv(const char *source, const char *thebot, char *buf)
{
    char *cmd, *s, *chan, *bmsg;
    int i;
    char orig[BUFSIZE];
    NickInfo *ni;
    strscpy (orig, buf, sizeof (orig));
    cmd = strtok(buf, " ");

    if (cmd)
            if (stricmp (cmd, "\1PING") == 0) {
                if (!(s = strtok (NULL, "")))
                    s = "\1";
                notice(thebot, source, "\1PING %s", s);

            } else if (stricmp (cmd, "\1VERSION\1") == 0) {
                if (!(s = strtok (NULL, "")))
                    s = "\1";
             notice(thebot, source, "%sVERSION Auspice IRC Services %s", s, s);
	
           } else if (!is_services_admin(source)) {
        	chan = strtok(NULL, "");
	    	for (i = 1; (i <= servbot) && (stricmp(botname[i],thebot) != 0); i++)
		;
            
	    	ni = findnick(botmaster[i]);

	    	if (ni && finduser(botmaster[i]))
            	if ((ni->flags & NI_IDENTIFIED) && (ni->flags & NI_BOTMSG)) {
	        	send_cmd(thebot, "PRIVMSG %s :(%s) %s %s", ni->nick, source, cmd, (chan)?chan:"");
            	}
	    	return;
          
            } else if (stricmp(cmd, "JOIN") == 0) {
                chan = strtok(NULL, " ");
                if (!chan) {
                        notice(thebot, source, "Syntax: JOIN \2Channel\2");
                } else {
			if (is_botowner(source, thebot) || is_services_root(source))
	                        send_cmd (thebot, "JOIN %s", chan);
			else
	                        notice(thebot, source, "Sorry you are not my master.");		
                }
            } else if (stricmp(cmd, "HELP") == 0) {
		if (is_botowner(source, thebot) || is_services_root(source)) {
		notice(thebot, source, "- JOIN      Join channel");	
		notice(thebot, source, "- PART      Part channel");	
		notice(thebot, source, "- SAY       Say in channel");	
		notice(thebot, source, "- ACT       Act in channel");	
		notice(thebot, source, "- QUIT      Quit the Bot");	
	        notice(thebot, source, "\2ENF OF HELP\2");
		} else
	                notice(thebot, source, "Sorry you are not my master.");		
 

            } else if (stricmp(cmd, "SAY") == 0) {
                chan = strtok(NULL, " ");
		bmsg = strtok(NULL, "");
                if (!chan || !bmsg) {
                        notice(thebot, source, "Syntax: SAY \2Channel\2 \2Message\2");
                } else {
			if (is_botowner(source, thebot) || is_services_root(source))
			     if (finduser(chan) || findchan(chan))
			     	send_cmd (thebot, "PRIVMSG %s :%s", chan, bmsg);
			     else
	                        notice(thebot, source, "Can not send to non exist channel or any services bots.");					
			else
	                        notice(thebot, source, "Sorry you are not my master.");		
                }

            } else if (stricmp(cmd, "ACT") == 0) {
                chan = strtok(NULL, " ");
		bmsg = strtok(NULL, "");
                if (!chan || !bmsg) {
                        notice(thebot, source, "Syntax: ACT \2Channel\2 \2Message\2");
                } else {
			if (is_botowner(source, thebot) || is_services_root(source))
			     if (finduser(chan) || findchan(chan))
			     send_cmd (thebot, "PRIVMSG %s :ACTION %s",chan, bmsg);
			     else
	                        notice(thebot, source, "Can not send to non exist channel or any services bots.");					

			else
	                        notice(thebot, source, "Sorry you are not my master.");		
                }

            } else if (stricmp(cmd, "PART") == 0) {
	        ChannelInfo *ci;
                chan = strtok(NULL, " ");	
		ci = cs_findchan(chan);
                if (!chan || !ci) {
                        notice(thebot, source, "Syntax: PART \2Channel\2");
		} else if (ci && (ci->bot) && (stricmp(ci->bot,thebot)==0)) {
                        notice(thebot, source, "Sorry, this bot have been assigned for this channel already.");
                } else {
			if (is_botowner(source, thebot) || is_services_root(source))
	                        send_cmd (thebot, "PART %s", chan);
			else
	                        notice(thebot, source, "Sorry you are not my master.");		
                }
            } else if (stricmp(cmd, "QUIT") == 0) {
		if (is_botowner(source, thebot) || is_services_root(source)) {
		        do_expireabot(thebot);
			quitbot(source, thebot, 0);
		} else
	                notice(thebot, source, "Sorry you are not my master.");		
            }
}
/*****************************************************************/
void statserv(const char *source, char *buf)
{
    char *cmd, *s, *chan;
    char orig[BUFSIZE];
    strscpy (orig, buf, sizeof (orig));
    cmd = strtok(buf, " ");

    if (cmd) 
	    if (stricmp (cmd, "\1PING") == 0)
	    {
	        if (!(s = strtok (NULL, "")))
	            s = "\1";
	        notice(s_OperServ, source, "\1PING %s", s);

	    } else if (stricmp (cmd, "\1VERSION\1") == 0) {
	        if (!(s = strtok (NULL, "")))
        	    s = "\1";
	        notice(s_StatServ, source, "%sVERSION Auspice IRC Services %s", s, s);

	    } else if (stricmp(cmd, "JOIN") == 0) {

		if (!is_services_root(source)) {
		       	return;
		}
		chan = strtok(NULL, " ");
		if (!chan) {
			notice(s_StatServ, source, "Syntax: JOIN \2Channel\2");
		} else {
		        send_cmd (s_StatServ, "JOIN %s", chan);
		}

	    } else if (stricmp(cmd, "PART") == 0) {
		if (!is_services_root(source)) {
		       	return;
		}
		chan = strtok(NULL, " ");
		if (!chan) {
			notice(s_StatServ, source, "Syntax: PART \2Channel\2");
		} else {
		        send_cmd (s_StatServ, "PART %s", chan);
		}

	    } else if (stricmp(cmd, "STATS") == 0) {
		do_mem(source);
	    } else if (stricmp(cmd, "SERVER") == 0) {
		do_statss(source);
	    } else if (stricmp(cmd, "SERVERS") == 0) {
		do_stats_server(source);
	    } else if (stricmp(cmd, "HELP") == 0) {
		notice(s_StatServ, source,"- SERVERS    List all servers.");
		notice(s_StatServ, source,"- STATS      Show Services Db Record.");
		if (is_services_oper(source)) {
		notice(s_StatServ, source,"- SERVER      List all servers with desc.");	
		}
		if (is_services_root(source)) {
		notice(s_StatServ, source,"- JOIN       Force StatServ join any channel.");	
		notice(s_StatServ, source,"- PART       Force StatServ leave any channel.");	
		}
	    }
}
/***********************************************************************/
int is_services_nick(const char *source)
{
	char *nick = sstrdup(source);
	if (is_servbot(nick))
		return 1;
	return 0;
}
/***********************************************************/
int is_servbot(char *nick)
{
      if ((stricmp(nick, s_OperServ) == 0)
	   || (stricmp(nick, s_RootServ) == 0)
	   || (stricmp(nick, s_NickServ) == 0)
	   || (stricmp(nick, s_MemoServ) == 0)
	   || (stricmp(nick, s_ChanServ) == 0)
	   || (stricmp(nick, "ChanOP") == 0)
	   || (stricmp(nick, s_HelpServ) == 0)
	   || (stricmp(nick, s_Agent) == 0)
	   || (stricmp(nick, s_ServicesX) == 0)
#ifdef DEFHOSTSERV
	   || (stricmp(nick, s_HostServ) == 0)
#endif
	   || (stricmp(nick, s_MassServ) == 0)
	   || (stricmp(nick, s_BotServ) == 0)
	   || (stricmp(nick, s_ServicesW) == 0)
	   || (stricmp(nick, s_StatServ) == 0)
	   || (stricmp(nick, s_SpamChecker) == 0)
	   || (check_ifbot(nick) == 1))
		return 1;
	return 0;
		
}
/***********************************************************************/
#ifdef DEFHOSTSERV
void HostServ(const char *source, char *buf)
{
    char *cmd;
    char *s, *chan, *nick, *host, *info;
    int i, j;
    User *u = finduser(source);
    NickInfo *ni;

    cmd = strtok(buf, " ");

    if (!cmd) {
	return;
    } else if (stricmp(cmd, "HELP") == 0) {
	if (is_services_helpop(source))
		notice(s_HostServ, source, "- SETHOST   Change you hostname");
	if (is_services_oper(source)) {
		notice(s_HostServ, source, "- ADD       Add auto sethost to a nick.");	
		notice(s_HostServ, source, "- DEL       Delete auto sethost from a nick.");	
		notice(s_HostServ, source, "- LIST      show nick list of auto sethost.");	
		notice(s_HostServ, source, "- CHANGE    Modify nickname host.");
	}
    } else if (stricmp(cmd, "\1PING") == 0) {
        if (!(s = strtok(NULL, "")))
            s = "\1";
        notice(s_OperServ, source, "\1PING %s", s);

    } else if (stricmp(cmd, "JOIN") == 0) {

	if (!is_services_root(source)) {
       		notice(s_HostServ, source, "Permision Denied.");
	       	return;
	}
	chan = strtok(NULL, " ");
	if (!chan) {
		notice(s_HostServ, source, "Syntax: JOIN \2Channel\2");
	} else {
	        send_cmd (s_HostServ, "JOIN %s", chan);
        	send_cmd (s_HostServ, "MODE %s +a %s", chan, s_ServicesW);
	}

    } else if (stricmp(cmd, "PART") == 0) {

	if (!is_services_root(source)) {
       		notice(s_HostServ, source, "Permision Denied.");
	       	return;
	}
	chan = strtok(NULL, " ");
	if (!chan) {
		notice(s_HostServ, source, "Syntax: PART \2Channel\2");
	} else {
	        send_cmd (s_HostServ, "PART %s", chan);
	}

   } else if (stricmp(cmd, "LIST") == 0) {
        s = strtok(NULL, " ");
	if (!is_services_oper(source)) {
       		notice(s_HostServ, source, "Permision Denied.");
	       	return;
	}
        if (!s)
            s = "*";
        slog("HO L %s (%s!%s@%s)", u->nick, u->nick, u->username, u->host);
        notice(s_HostServ, source, "\2Current Host list:\2");
        for (i = 0; i < nhostline; i++) {
            if (!s || match_wild(s, s_hlines[i].nick)) {

              if (ni = findnick(s_hlines[i].nick))
                notice(s_HostServ, source, " %d) %-16s (%s) added by %s",
                                i+1, s_hlines[i].nick, s_hlines[i].host, s_hlines[i].who);
              else
                notice(s_HostServ, source, " %d) %-16s (Not Registered)",
                                i+1, s_hlines[i].nick);
            }
        }
        notice(s_HostServ,source,"\2End of List\2");

    	} else if (stricmp(cmd,"ADD") == 0) {
        nick = strtok(NULL, " ");
        host = strtok(NULL, "");
        if (!nick || !host) {
            notice(s_HostServ, source, "Syntax: \2ADD\2 <nick> <host>");
            return;
        }
        if (!findnick(nick)) {
                notice(s_HostServ, source, "Nick %s isn't registered!", nick);
                return;
        }

	if (!is_services_oper(source)) {
       		notice(s_HostServ, source, "Permision Denied.");
	       	return;
	}

        for (i = 0; i < nhostline; i++) {
            if (stricmp(s_hlines[i].nick,nick) == 0) {
                notice(s_HostServ,source,"%s is already present "
                    "on the Host list.", nick);
                return;
            }
        }

        if (nhostline >= hostline_size) {
            if (hostline_size < 8)
                hostline_size = 8;
            else
                hostline_size *= 2;
            s_hlines = srealloc(s_hlines, sizeof(*s_hlines) *hostline_size);
        }

        if (!nhostline || (nhostline < MAX_HOSTSERV) || (MAX_HOSTSERV == 0)) {
            s_hlines[nhostline].nick = sstrdup(nick);
            s_hlines[nhostline].host = sstrdup(host);
            strscpy(s_hlines[nhostline].who, source, NICKMAX);

            notice(s_HostServ, source, "Added nick %s into auto change host list (%s).",nick,host);
	    show_next_db(source, s_ServicesW);
            slog("HO A %s %s by %s", nick, host, source);
            ++nhostline;
            if (readonly) {
                notice(s_HostServ, source,
                    "\2Notice:\2 Changes will not be saved! "
                    " Services is in read-only mode.");
            }
        }
        else {
            notice(s_HostServ,source,"Sorry, Host list is full.");
            wallops(SERVER_NAME,"\2Warning\2: host(nick) list is full. (%d)", MAX_HOSTSERV);
        }

    } else if (stricmp(cmd, "CHANGE") == 0) {
        int i;
        nick = strtok(NULL, " ");
        info = strtok(NULL, "");

        if (!is_services_oper(source)) {
            notice(s_HostServ, source, "Permission denied.");
            return;
        }

        if ((nick == NULL) || (info == NULL)) {
          notice(s_HostServ, source, "Syntax: \2CHANGE\2 <nick/number> <newhost>");
          return;
	}
            if (strspn(nick, "1234567890") == strlen(nick) &&
                                (i = atoi(nick)) > 0 && i <= nhostline) {

                free(s_hlines[i-1].host);
                s_hlines[i-1].host = sstrdup(info);
	        strscpy(s_hlines[i-1].who, source, NICKMAX);
                notice(s_HostServ, source, "Nick %s auto change host to %s",
                             s_hlines[i-1].nick, sstrdup(info));
                return;
            }

         for (i = 0; i < nhostline; i++) {
               if (stricmp(s_hlines[i].nick, nick) == 0) {

                   free(s_hlines[i].host);
                   s_hlines[i].host = sstrdup(info);
	           strscpy(s_hlines[i].who, source, NICKMAX);
	           notice(s_HostServ, source, "Nick %s auto change host to %s",
                             s_hlines[i].nick, sstrdup(info));
               }
         }

    } else if (stricmp(cmd, "DEL") == 0) {
        int i;
	if (!is_services_oper(source)) {
       		notice(s_HostServ, source, "Permision Denied.");
	       	return;
	}
        if (nick = strtok(NULL, " ")) {
            if (strspn(nick, "1234567890") == strlen(nick) &&
                                (i = atoi(nick)) > 0 && i <= nhostline) {
                strcpy(nick,s_hlines[i-1].nick);
            }
            if (del_hostline(nick)) {
                notice(s_HostServ, source, "%s removed from Host list.", nick);
                slog("HO D %s (%s!%s@%s)",
                nick, u->nick, u->username, u->host);

                if (readonly) {
                    notice(s_HostServ, source,
                        "\2Notice:\2 Changes will not be saved!  Services is in readonly mode.");
                }
            } else {
                notice(s_HostServ, source, "%s not found on Host list.",
                    nick);
            }
        } else {
            notice(s_HostServ, source, "Syntax: \2DEL\2 <nick>");
        }

    } else if (stricmp(cmd, "SETHOST") == 0) {
	if (!is_services_helpop(source)) {
       		notice(s_HostServ, source, "Permision Denied.");
	       	return;
	}
	chan = strtok(NULL, " ");  /* Chan = Host */
	if (!chan) {
		notice(s_HostServ, source, "Syntax: SETHOST \2<host>\2");
	} else {
	        send_cmd (s_HostServ, "CHGHOST %s %s", source, chan);
		notice(s_HostServ, source, "Your host name set to \2%s\2", chan);
		notice(s_HostServ, source, "To turn it off type /mode %s -x", source);
	        slog("OPER HOST %s (%s!%s@%s) %s", u->nick, u->nick, u->username, u->host, chan);
	}
   }
}
#endif
/***********************************************************************/
void massserv(const char *source, char *buf)
{
    char *cmd, *s, *chan;
    char orig[BUFSIZE];
    strscpy (orig, buf, sizeof (orig));
    cmd = strtok(buf, " ");

    if (cmd) 
	    if (stricmp (cmd, "\1PING") == 0)
	    {
	        if (!(s = strtok (NULL, "")))
	            s = "\1";
	        notice(s_MassServ, source, "\1PING %s", s);

	    } else if (stricmp (cmd, "\1VERSION\1") == 0) {
	        if (!(s = strtok (NULL, "")))
        	    s = "\1";
	        notice(s_MassServ, source, "%sVERSION Auspice IRC Services %s", s, s);

	    } else if (!is_services_root(source)) {
		notice(s_MassServ, source, "Permission denied.");
	       	return;
	    } else if (stricmp(cmd, "JOIN") == 0) {

		if (!is_services_root(source)) {
			notice(s_MassServ, source, "Permission denied.");
		       	return;
		}
		chan = strtok(NULL, " ");
		if (!chan) {
			notice(s_MassServ, source, "Syntax: JOIN \2Channel\2");
		} else {
		        send_cmd (s_MassServ, "JOIN %s", chan);
		}

	    } else if (stricmp(cmd, "PART") == 0) {

		if (!is_services_root(source)) {
			notice(s_MassServ, source, "Permission denied.");
		       	return;
		}
		chan = strtok(NULL, " ");
		if (!chan) {
			notice(s_MassServ, source, "Syntax: PART \2Channel\2");
		} else {
		        send_cmd (s_MassServ, "PART %s", chan);
		}

            } else if (stricmp(cmd, "SERVJOIN") == 0) {
	        chan = strtok(NULL, " ");
		if (!is_services_root(source)) {
			notice(s_MassServ, source, "Permission denied.");
		       	return;
		}
	        if (!chan) {
		    notice(s_MassServ, source, "Syntax: SERVJOIN <channel>");
	            return;
		}
		send_cmd(s_OperServ, "JOIN %s", chan);
		send_cmd(s_OperServ, "MODE %s +o %s", chan, s_OperServ);
		send_cmd(s_RootServ, "JOIN %s", chan);
	        send_cmd(s_RootServ, "MODE %s +o %s", chan, s_RootServ);
	        send_cmd(s_NickServ, "JOIN %s", chan);
	        send_cmd(s_NickServ, "MODE %s +o %s", chan, s_NickServ);
	        send_cmd(s_ChanServ, "JOIN %s", chan);
	        send_cmd(s_ChanServ, "MODE %s +o %s", chan, s_ChanServ);
	        send_cmd(s_MemoServ, "JOIN %s", chan);
	        send_cmd(s_MemoServ, "MODE %s +o %s", chan, s_MemoServ);
	        send_cmd(s_HelpServ, "JOIN %s", chan);
	        send_cmd(s_HelpServ, "MODE %s +o %s", chan, s_HelpServ);
	        send_cmd(s_Agent, "JOIN %s", chan);
	        send_cmd(s_Agent, "MODE %s +o %s", chan, s_Agent);
	        send_cmd(s_ServicesX, "JOIN %s", chan);
	        send_cmd(s_ServicesX, "MODE %s +o %s", chan, s_ServicesX);
	        send_cmd(s_ServicesW, "JOIN %s", chan);
	        send_cmd(s_ServicesW, "MODE %s +o %s", chan, s_ServicesW);
	        send_cmd(s_GlobalNoticer, "JOIN %s", chan);
	        send_cmd(s_GlobalNoticer, "MODE %s +o %s", chan, s_GlobalNoticer);
	        send_cmd(s_SpamChecker, "JOIN %s", chan);
	        send_cmd(s_SpamChecker, "MODE %s +o %s", chan, s_SpamChecker);
	        send_cmd(s_StatServ, "JOIN %s", chan);
	        send_cmd(s_StatServ, "MODE %s +o %s", chan, s_StatServ);
	        send_cmd(s_MassServ, "JOIN %s", chan);
	        send_cmd(s_MassServ, "MODE %s +o %s", chan, s_MassServ);
	        send_cmd(s_BotServ, "JOIN %s", chan);
	        send_cmd(s_BotServ, "MODE %s +o %s", chan, s_BotServ);
	        send_cmd(s_HostServ, "JOIN %s", chan);
	        send_cmd(s_HostServ, "MODE %s +o %s", chan, s_HostServ);
            } else if (stricmp(cmd, "SERVPART") == 0) {
	        chan = strtok(NULL, " ");
		if (!is_services_root(source)) {
			notice(s_MassServ, source, "Permission denied.");
		       	return;
		}
	        if (!chan) {
		    notice(s_MassServ, source, "Syntax: SERVPART <channel>");
	            return;
		}
	        send_cmd(s_OperServ, "PART %s", chan);
	        send_cmd(s_RootServ, "PART %s", chan);
	        send_cmd(s_NickServ, "PART %s", chan);
	        send_cmd(s_ChanServ, "PART %s", chan);
	        send_cmd(s_MemoServ, "PART %s", chan);
	        send_cmd(s_HelpServ, "PART %s", chan);
	        send_cmd(s_Agent, "PART %s", chan);
	        send_cmd(s_ServicesX, "PART %s", chan);
	        send_cmd(s_ServicesW, "PART %s", chan);
	        send_cmd(s_GlobalNoticer, "PART %s", chan);
	        send_cmd(s_SpamChecker, "PART %s", chan);
	        send_cmd(s_StatServ, "PART %s", chan);
	        send_cmd(s_BotServ, "PART %s", chan);
	        send_cmd(s_MassServ, "PART %s", chan);
	        send_cmd(s_HostServ, "PART %s", chan);

	    } else if (stricmp(cmd, "ALLBOTJOIN") == 0 ) {
		int i;
	        chan = strtok(NULL, " ");
		if (!is_services_root(source)) {
			notice(s_MassServ, source, "Permission denied.");
		       	return;
		}
	        if (!chan) {
		    notice(s_MassServ, source, "Syntax: ALLBOTJOIN <channel>");
	            return;
		}
		for (i=1; i<=servbot; i++) {
		        send_cmd(botname[i], "JOIN %s", chan);
			send_cmd(botname[i], "MODE %s +o %s", chan, botname[i]);

		}

	    } else if (stricmp(cmd, "ALLBOTPART") == 0 ) {
		int i;
	        chan = strtok(NULL, " ");
		if (!is_services_root(source)) {
			notice(s_MassServ, source, "Permission denied.");
		       	return;
		}
	        if (!chan) {
		    notice(s_MassServ, source, "Syntax: ALLBOTPART <channel>");
	            return;
		}
		for (i=1; i<=servbot; i++)
		        send_cmd(botname[i], "PART %s", chan);

            } else if (stricmp(cmd, "ALLBOTQUIT") == 0) {
		int i;
		if (!is_services_coder(source)) {
			notice(s_MassServ, source, "Permission denied.");
		       	return;
		}
	        wallops(s_BotServ, "All Services bot quited by \2%s\2", source);
		for (i=1; i<=servbot; i++) {
	                send_cmd (botname[i], "QUIT :Quit by %s", source);
			botname[i] = NULL;
			botmaster[i] = NULL;
		}
		servbot = 0;
            } else if (stricmp(cmd, "ALLJOIN") == 0) {
	        chan = strtok(NULL, " ");
		if (!is_services_coder(source)) {
			notice(s_MassServ, source, "Permission denied.");
		       	return;
		}
	        if (!chan) {
		    notice(s_MassServ, source, "Syntax: ALLJOIN <channel>");
	            return;
		}
		alljoin(chan);
	    } else if (stricmp(cmd, "HELP") == 0) {
		notice(s_MassServ, source,"Mass Services still under construction.");
		notice(s_MassServ, source,"- SERVJOIN      Make all Services bots Join a channel");
		notice(s_MassServ, source,"- SERVPART      Make all Services bots Part a channel");
		notice(s_MassServ, source,"- ALLBOTJOIN    Make all bots Join a channel");
		notice(s_MassServ, source,"- ALLBOTPART    Make all bots Part a channel");		
		notice(s_MassServ, source,"- ALLBOTQUIT    Exit all bots");
		notice(s_MassServ, source,"- ALLJOIN       Make all users join a channel");
	    } else
		notice(s_MassServ, source,"Mass Services still under construction.");
}
/************************************************************************/
void createbot(const char *source, char *nick, int temps)
{
#ifdef BAHAMUT
        send_cmd (NULL, "NICK %s 1 %lu %s %s %s %s %lu:%s",
        nick, time(NULL), "+", "ServerBot", services_host, server_name, time(NULL), "Services Bot");
#elif defined(NICKV2)
        send_cmd (NULL, "NICK %s 1 %ld %s %s %s 1 + * :%s",
        nick, time(NULL), "ServerBot", services_host, server_name, "Services Bot");
#else
	send_cmd (NULL, "NICK %s 1 %lu %s %s %s %lu:%s",
        nick, time(NULL), "ServerBot", services_host, server_name, time(NULL), "Services Bot");
#endif
        send_cmd(nick, "MODE %s +%s", nick, BOTMODE);
        send_cmd(nick, "JOIN %s", snoopchan);
        send_cmd(nick,"SWHOIS %s :is a Services Bot",nick);
	if (!temps)
        wallops(s_BotServ, "Services bot \2%s\2 connected by \2%s\2", nick, source);
	++servbot;
	botname[servbot] = sstrdup(nick);
	if (temps) {
		botmaster[servbot] = sstrdup("ChanOP");
	} else {
		botmaster[servbot] = sstrdup(source);
	}
}
/************************************************************************/
void quitbot(const char *source, const char *nick, int temps)
{
	int i, j;

        send_cmd (nick, "QUIT :Quit by %s", source);
	if(!temps)
	        wallops(s_BotServ, "Services bot \2%s\2 quited by \2%s\2", nick, source);
        for (i=1; i<=servbot; i++)
	if (stricmp(botname[i], nick) == 0)
		break;
	for (j=i; j<=(servbot-1) ; ++j) {
		strcpy(botname[j], botname[j+1]);
		strcpy(botmaster[j], botmaster[j+1]);
	}
	botname[servbot] = NULL;
	botmaster[servbot] = NULL;
	--servbot;
}
/************************************************************************/
void sbotserv(const char *source, char *buf)
{
    char *cmd, *s, *chan, *bchan;
    char orig[BUFSIZE];
    strscpy (orig, buf, sizeof (orig));
    cmd = strtok(buf, " ");
//    log(buf);
    if (cmd) 
	    if (stricmp (cmd, "\1PING") == 0)
	    {
	        if (!(s = strtok (NULL, "")))
	            s = "\1";
	        notice(s_BotServ, source, "\1PING %s", s);

	    } else if (stricmp (cmd, "\1VERSION\1") == 0) {
	        if (!(s = strtok (NULL, "")))
        	    s = "\1";
	        notice(s_BotServ, source, "%sVERSION Auspice IRC Services %s", s, s);

	    } else if (stricmp(cmd, "LIST") == 0) {
		int i;
	        notice(s_BotServ, source, "\2%d BOT(s) IN LIST\2", servbot);
		for(i=1; i <= servbot; i++)
		        notice(s_BotServ, source, "%d)- %s (%s)",i, botname[i], botmaster[i]);
	        notice(s_BotServ, source, "\2ENF OF LIST\2");

	    } else if (!is_services_admin(source)) {

       		notice(s_BotServ, source, "Permision Denied.");
	       	return;

	    } else if (stricmp(cmd, "JOIN") == 0) {

		if (!is_services_admin(source)) {
	       		notice(s_BotServ, source, "Permision Denied.");
		       	return;
		}
		chan = strtok(NULL, " ");
		if (!chan) {
			notice(s_BotServ, source, "Syntax: JOIN \2Channel\2");
		} else {
		        send_cmd (s_BotServ, "JOIN %s", chan);
		}

	    } else if (stricmp(cmd, "PART") == 0) {
		if (!is_services_admin(source)) {
	       		notice(s_BotServ, source, "Permision Denied.");
		       	return;
		}
		chan = strtok(NULL, " ");
		if (!chan) {
			notice(s_BotServ, source, "Syntax: PART \2Channel\2");
		} else {
		        send_cmd (s_BotServ, "PART %s", chan);
		}
	    } else if (stricmp(cmd, "WHOIS") == 0) {
		if (!is_services_admin(source)) {
	       		notice(s_BotServ, source, "Permision Denied.");
		       	return;
		}
		chan = strtok(NULL, " ");
		if (!chan) {
			notice(s_BotServ, source, "Syntax: WHOIS \2Botnick\2");
		} else {
			int i;
			for (i=1; i<=servbot; i++)
			if (stricmp(chan, botname[i]) == 0) {
				notice(s_BotServ, source,"WHOIS: Bot %s connected by %s", botname[i], botmaster[i]);
				return;
			}
		}

	    } else if (stricmp(cmd, "SNOOP") == 0) {
		if (!is_services_root(source)) {
	       		notice(s_BotServ, source, "Permision Denied.");
		       	return;
		}
		chan = strtok(NULL, " ");
		bchan = strtok(NULL, " ");
		if (chan && (stricmp(chan,"NONE") == 0)) {
			aspychan = 0;
			notice(s_BotServ, source, "Disable Channel Spy feature");
		} else if (!chan || !bchan) {
			notice(s_BotServ, source, "Syntax: SNOOP \2source\2 \2dest\2");
		} else if (chan && !bchan) {
			notice(s_BotServ, source, "Syntax: SNOOP \2source\2 \2dest\2");
		} else {
			asource = sstrdup(chan);
			adest = sstrdup(bchan);
			aspychan = 1;
			notice(s_BotServ, source, "All message in channel %s forward to %s.", chan, bchan);		
		}

	    } else if (stricmp(cmd, "HTM") == 0) {
		int i;
		for(i=1; i <= servbot; i++)
		        send_cmd(botname[i], "MODE %s +d", botname[i]);
	        notice(s_BotServ, source, "\2All bot now mute.\2");
	    } else if (stricmp(cmd, "NHTM") == 0) {
		int i;
		for(i=1; i <= servbot; i++)
		        send_cmd(botname[i], "MODE %s -d", botname[i]);
	        notice(s_BotServ, source, "\2All now listend to channel command.\2");
	    } else if (stricmp(cmd, "CONNECT") == 0) {
		chan = strtok(NULL, " ");
		if (!is_services_admin(source)) {
	       		notice(s_BotServ, source, "Permision Denied.");
		       	return;
		}
		if (!chan) {
			notice(s_BotServ, source, "Syntax: CONNECT \2nick\2");
		} else if (is_validnick(chan)) {
			if (finduser(chan) || (is_servbot(chan) == 1)) {
				notice(s_BotServ, source, "Current nick is using.");
			} else {
				createbot(source, chan, 0);
			}
		} else {
	       		notice(s_BotServ, source, "ERROR: Invalid bot nick.");
		}

	    } else if (stricmp(cmd, "HELP") == 0) {
		notice(s_BotServ, source,"Bot Services still under construction.");
		notice(s_BotServ, source, "- CONNECT Connect Services Bot.");
		notice(s_BotServ, source, "- LIST    List all Services Bot.");
		notice(s_BotServ, source, "- WHOIS   Do whois a bot.");
		notice(s_BotServ, source, "- SNOOP   Spy channel.");
		notice(s_BotServ, source, "- HTM     Make bot not listend to channel messages.");
		notice(s_BotServ, source, "- NHTM    Make bot listend to channel messages.");
	    } else
		notice(s_BotServ, source,"Bot Services still under construction.");
}
/********************************************************************/
void bmsgchan(const char *source, char *chan, char *buf)
{
	int cha;
	ChannelInfo *ci;
	char *cmd, *cmd2, *cmd3;
	User *u = finduser(source);
	Channel *c = findchan(chan);
        char *nick = sstrdup(source);
	int i = 0;
        char *s = buf;

	ci = cs_findchan(chan);

        if (ci && (ci->flags & CI_CLOSED))
		return;

	if ((aspychan == 1) && (stricmp(chan,asource) == 0))
	     send_cmd(s_OperServ, "PRIVMSG %s :<%s/%s> %s", adest, source, chan, buf);
	if (!ci || !u || !c) return;

	cmd2 = sstrdup(buf);
	cmd = strtok(cmd2, " ");

	if (source && cmd && (stricmp(cmd, "`FUN")==0)) {
		cmd3 = strtok(NULL, " ");
		if (!cmd3) {
			notice(s_ChanServ, source,"SYNTAX for fun: <cleanoffice|milk|milo|squish>");
		} else if (stricmp(cmd3, "CLEANOFFICE")==0) {
			send_cmd (s_ChanServ, "PRIVMSG %s :Nick off %s you lazy loser, Do it ya self", chan, source);
		} else if (stricmp(cmd3, "MILK")==0) {
			send_cmd (s_ChanServ, "PRIVMSG %s :ACTION pours %s a long glass of milk. huhuhuhuhu milk", chan, source);
		} else if (stricmp(cmd3, "MILO")==0) {
			send_cmd (s_ChanServ, "PRIVMSG %s :ahhh milo wot a good idea", chan);
			send_cmd (s_ChanServ, "PRIVMSG %s :ACTION decides hot milo made with milk and marshmellows sounds great", chan);
			send_cmd (s_ChanServ, "PRIVMSG %s :OK who wants one?????", chan);
			send_cmd (s_ChanServ, "PRIVMSG %s :aren't ya glad you have a clever bot ;)", chan);
		} else if (stricmp(cmd3, "SQUISH")==0) {
			send_cmd (s_ChanServ, "PRIVMSG %s :ACTION squishes %s", chan, source);
		}
	}

	if (ci && !(ci->bot)) {
		return;
	} else {

	if (!((ci->botflag & CBI_PROTOP) && (is_voiced(source,chan) || is_chanop(source, chan) 
#ifdef DEFHALFOP
|| is_chanhop(source, chan) 
#endif
#ifdef DEFPROTECT
|| is_prots(source, chan)
#endif
))) {
		/***************** CAP *******************/
		if ((ci->botflag & CBI_KICKCAP) && (cha = strlen(buf) > 10)) {
		        do {
		        	if (isupper(*s)) i++;
			} while (*s++);
			cha = strlen(buf);
			if (i >= 10 && i*100/cha>=25)
			{
				if (
#ifdef DEFHALFOP
is_chanhop(source, ci->name) || 
#endif
				is_chanop(source, ci->name)) {
					notice(ci->bot, source,"Your caplocks excess the limit please drop it.");
					dropmode(nick, ci);
					return;
				} else {
					botkick(nick, ci, "Please turn caps lock OFF!");
					return;
				}
			}
		}
		/********** END CAP **************/

                /********** BOLDS ****************/
		if ((ci->botflag & CBI_KICKBOLDS) && strchr(buf, 2)) {
			if (
#ifdef DEFHALFOP
is_chanhop(source, ci->name) || 
#endif
			is_chanop(source, ci->name)) {
				notice(ci->bot, source,"Don't use bolds on this channel!.");
				dropmode(nick, ci);
				return;
			} else {
				botkick(nick, ci, "Don't use bolds on this channel!");
				return;
			}
		}
                /***************** END BOLD **************/
		/***************** COLORS ****************/
		if ((ci->botflag & CBI_KICKCOLORS) && strchr(buf, 3)) {
			if (
#ifdef DEFHALFOP
is_chanhop(source, ci->name) || 
#endif
			is_chanop(source, ci->name)) {
				notice(ci->bot, source,"Don't use colors on this channel!");
				dropmode(nick, ci);
				return;
			} else {
				botkick(nick, ci, "Don't use colors on this channel!");
				return;
			}
		}
                /******************** END COLOR ******************/           
		/******************** REVERSES *******************/
		if ((ci->botflag & CBI_KICKREVERSES) && strchr(buf, 22)) {
			if (
#ifdef DEFHALFOP
is_chanhop(source, ci->name) || 
#endif
				is_chanop(source, ci->name)) {
				notice(ci->bot, source,"Don't use reverses on this channel!");
				dropmode(nick, ci);
				return;
			} else {
				botkick(nick, ci, "Don't use reverses on this channel!");
				return;
			}
		}
		/******************** END REVERSES *******************/
		/******************** UNDERLINES *********************/
		if ((ci->botflag & CBI_KICKUNDERLINES) && strchr(buf, 31)) {
			if (
#ifdef DEFHALFOP
is_chanhop(source, ci->name) || 
#endif
			is_chanop(source, ci->name)) {
				notice(ci->bot, source,"Don't use underlines on this channel!");
				dropmode(nick, ci);
				return;
			} else {
				botkick(nick, ci, "Don't use underlines on this channel!");
				return;
			}
		}
		/******************* END UNDERLINES ******************/
		if ((ci->botflag & CBI_KICKBADWORDS) && is_badwords(buf)) {
			if (
#ifdef DEFHALFOP
is_chanhop(source, ci->name) || 
#endif
			is_chanop(source, ci->name)) {
				notice(ci->bot, source,"Don't use badword!!");
				dropmode(nick, ci);
				return;
			} else {
				botkick(nick, ci, "Don't use bad word on this channel!!");
				return;
			}
		}
	}	
		/******************** BOT COMMAND ****************/
		if (ci->botflag & CBI_BOTCOM) {
			abotcom(source, ci, buf);
		}
	}
}
/************************************************************************/
void botkick(char *nick, ChannelInfo *ci, char *buf)
{
	if (is_on_chan(nick, ci->name) && ci->bot) {
	       User *u2 = finduser(nick);
	       char *av[3];
               av[0] = sstrdup(ci->name);
               av[1] = sstrdup("+b");
	       av[2] = bcreate_mask(u2, 1);
			
	       send_cmd(ci->bot, "MODE %s +b %s  %lu", ci->name, av[2], time(NULL));
	       do_cmode(ci->bot, 3, av);

                av[1] = sstrdup(nick);
		av[2] = sstrdup(buf);
		send_cmd(ci->bot, "KICK %s %s :%s", av[0], av[1], av[2]);
		do_kick(ci->bot, 3, av);
                free(av[2]);
                free(av[1]);
                free(av[0]);
	}		
}
/************************************************************************/
void dropmode(char *nick, ChannelInfo *ci)
{
	if (is_on_chan(nick, ci->name)) {
		if (is_chanop(nick, ci->name))
		change_cmode (ci->bot, ci->name, "-o", nick);
#ifdef DEFPROTECT
		if (is_prots(nick, ci->name))
		change_cmode (ci->bot, ci->name, "-a", nick);
#endif
#ifdef DEFHALFOP
		if (is_chanhop(nick, ci->name))
		change_cmode (ci->bot, ci->name, "-h", nick);
#endif
		if (is_voiced(nick, ci->name))
		change_cmode (ci->bot, ci->name, "-v", nick);
	}		
}
/************************************************************************/
static char *revealmask(char *word1, char *word2, char *word3) {
     char *wordss;
     char *wordsss;
     int i;
     wordsss = sstrdup(word1);
     wordss = wordsss;
     for (i=1; i <= strlen(word1); i++)
	if ((*wordss == toupper(*word2)) || (*wordss == *word3)) {
             *wordss++;
             *word3++;
	} else {
             *word3++;
             *wordss++ = '*';
	}
     *wordss = 0;
     return wordsss;
}
/************************************************************************/
static void abotcom(const char *source, ChannelInfo *ci, char *buf)
{
   char *chan = ci->name;
   User *u = finduser (source);
   char *cmd, *param;
   cmd = strtok(buf, " ");
		
   if(u && cmd) {
        /******************** OP Command ********************/

	if ((stricmp(cmd,"!op")==0) || (stricmp(cmd,"!up")==0) || (stricmp(cmd,"`up")==0)) {
	   param = strtok(NULL, " ");

	   if (!is_chanop(source, chan) && (get_access(u, ci) < 5)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   } 

	  if (!param) {
		if (!is_on_chan(source, chan))
			notice(ci->bot, source,"You are not in channel %s.", chan);
		else if (is_chanop(source, chan))
			notice(ci->bot, source,"You already have op in %s.", chan);
		else
		        change_cmode (ci->bot, chan, "+o", source);

	  } else {
		User *u2 = finduser(param);
		if (!u2 || !is_on_chan(param, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if ((ci->flags & CI_OPGUARD) && (get_access(u2, ci) < 5))
			notice(ci->bot, source,"Channel %s set opguard on please respect founder.",chan);
		else if (is_chanop(param, chan))
			notice(ci->bot, source,"User %s already have op in %s.",param,chan);
		else
		        change_cmode (ci->bot, chan, "+o", param);
	  }

	}

       /********************* END OP *************************/
	else if (((stricmp(cmd,"`gnum")==0) || (stricmp(cmd,"!gnum")==0)) && (ci->botflag & CBI_GUESSNUM)){
	   int theguess;
	   param = strtok(NULL, " ");
	   if (!param) {
		notice(ci->bot, source,"Please Enter 4 digits number.");
		return;
	   } else if (strlen(param) > 4) {
		notice(ci->bot, source,"Please Enter 4 digits number.");
		return;
	   } else {
	        Channel *c = findchan(chan);
		theguess = atoi(param);
		if (theguess > c->ngames) {
		     send_cmd (ci->bot, "PRIVMSG %s :ACTION %s \2%d\2 is too big",chan,source, theguess);
		} else if (theguess == c->ngames) {
		     double x =  RAND_MAX + 1.0;
		     c->ngames = 1 + rand() * (999/x);
		     u->games++;
		     send_cmd (ci->bot, "PRIVMSG %s :ACTION \37%s\37 \2You Win\2, your point is %d",chan,source, u->games);
		     send_cmd (ci->bot, "NOTICE @%s :Answer is \2%d\2",chan,c->ngames);
	       	send_cmd (ci->bot, "NOTICE %s :New Game start, type \2`GNum <number>\2 to make guess!!",chan);

		} else {
		     send_cmd (ci->bot, "PRIVMSG %s :ACTION \37%s\37 \2%d\2 is too small",chan,source, theguess);
		}
	   } }
       /*****************************************************/
	else if (((stricmp(cmd,"`nextnum")==0) || (stricmp(cmd,"!nextnum")==0)) && (ci->botflag & CBI_GUESSNUM)){
	        Channel *c = findchan(chan);
	        double x =  RAND_MAX + 1.0;
		if (!is_chanop(source, chan)) {
			notice(ci->bot, source,"Only op can do.");
			return;
		}   
	        c->ngames = 1 + rand() * (999/x);
	        send_cmd (ci->bot, "NOTICE @%s :Answer is \2%d\2",chan,c->ngames);
	        send_cmd (ci->bot, "NOTICE %s :New Game start, type \2`GNum <number>\2 to make guess!!",chan);
	   }
       /*****************************************************/
	else if (((stricmp(cmd,"`gword")==0) || (stricmp(cmd,"!gword")==0)) && (ci->botflag & CBI_GUESSWRD)){
	   param = strtok(NULL, " ");
	   if (!param) {
		notice(ci->bot, source,"Syntax: `GWord <word or char>.");
		return;
	   } else if (strlen(param) > 50) {
		notice(ci->bot, source,"There are no any word longer than 50 chars.");
		return;
	   } else {
	        Channel *c = findchan(chan);
		if (c->wgames == 0) {
		     c->wgames = 1;
		     c->wgame = sstrdup(wordgame[c->wgames]);
		}
		if (stricmp(param, wordgame[c->wgames]) == 0) {
		     char *wordss;
		     char *wordsss;
		     int i;
		     double x =  RAND_MAX + 1.0;
		     c->wgames = 1 + rand() * (wgameline/x);
		     u->games++;
		     send_cmd (ci->bot, "PRIVMSG %s :ACTION %s \2You Win Guessing Word Game\2, your Score now is %d",chan,source, u->games);
		     send_cmd (ci->bot, "NOTICE %s :New Word Guessing Game start, type \2`GWord <word or char>\2 to make guess!!",chan);

		     wordss = sstrdup(wordgame[c->wgames]);
		     wordsss = wordss;
		     *wordss++;
		     for (i=2; i< strlen(wordgame[c->wgames]); i++)
			*wordss++ = '*';
		     c->wgame = sstrdup(wordsss);
		     send_cmd (ci->bot, "NOTICE %s :The new word is \2%s\2 Hints: \37%s\37.",chan,wordsss,wordhelp[c->wgames]);
		     send_cmd (ci->bot, "NOTICE @%s :Answer is \2%s\2",chan,wordgame[c->wgames]);
		} else {
		     c->wgame = sstrdup(revealmask(wordgame[c->wgames],param,c->wgame));
		     send_cmd (ci->bot, "PRIVMSG %s :ACTION The word now is \2%s\2 Hints: \37%s\37.",chan,c->wgame,wordhelp[c->wgames]);
		}
	   }
	}
	/*******************************************************/
	else if (((stricmp(cmd,"`nextword")==0) || (stricmp(cmd,"!nextword")==0)) && (ci->botflag & CBI_GUESSWRD)){
	        Channel *c = findchan(chan);
	        double x =  RAND_MAX + 1.0;
 	        char *wordss;
	        char *wordsss;
	        int i;
		if (
#ifdef DEFHALFOP
is_chanhop(source, chan) || 
#endif
		is_chanop(source,chan)) {
		     c->wgames = 1 + rand() * (wgameline/x);
		     send_cmd (ci->bot, "NOTICE %s :New Word Guessing Game start, type \2`GWord <word or char>\2 to make guess!!",chan);

		     wordss = sstrdup(wordgame[c->wgames]);
		     wordsss = wordss;
		     *wordss++;
		     for (i=2; i< strlen(wordgame[c->wgames]); i++)
			*wordss++ = '*';
		     c->wgame = sstrdup(wordsss);
		     send_cmd (ci->bot, "NOTICE %s :The new word is \2%s\2 Hints: \37%s\37.",chan,wordsss,wordhelp[c->wgames]);
		     send_cmd (ci->bot, "NOTICE @%s :Answer is \2%s\2",chan,wordgame[c->wgames]);
		} else {
			notice(ci->bot, source,"Only op can do.");
			return;
		}   
	   }
       /********************** DEOP *************************/
	else if ((stricmp(cmd,"!deop")==0) || (stricmp(cmd,"!down")==0) || (stricmp(cmd,"`deop")==0)) {
	   param = strtok(NULL, " ");
	   if (!is_chanop(source, chan) && (get_access(u, ci) < 5)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		if (!is_on_chan(source, chan))
			notice(ci->bot, source,"You are not in channel %s.", chan);
		else if (!is_chanop(source, chan))
			notice(ci->bot, source,"You already deop in %s.", chan);
		else
		        change_cmode (ci->bot, chan, "-o", source);

	  } else {
		if (!is_on_chan(param, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if (!is_chanop(param, chan))
			notice(ci->bot, source,"User %s already deop in %s.",param,chan);
		else
		        change_cmode (ci->bot, chan, "-o", param);
	  }
       }
         /*************************** End Deop *********************/
         /*************************** molest ***********************/
	else if (stricmp(cmd,"!molest")==0) {

	if (!is_on_chan(source, chan))
		notice(ci->bot, source,"You are not in channel %s.", chan);
	else
		dropmode(sstrdup(source), ci);

       }
         /*************************** End MoLest *******************/
         /*************************** Voice ************************/
	else if (stricmp(cmd,"!voice")==0) {
	   param = strtok(NULL, " ");

	   if (!is_chanop(source, chan) && (get_access(u, ci) < 4)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		if (!is_on_chan(source, chan))
			notice(ci->bot, source,"You are not in channel %s.", chan);
		else if (is_voiced(source, chan))
			notice(ci->bot, source,"You already have voice in %s.", chan);
		else
		        change_cmode (ci->bot, chan, "+v", source);

	  } else {
		if (!is_on_chan(param, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if (is_voiced(param, chan))
			notice(ci->bot, source,"User %s already have voice in %s.",param,chan);
		else
		        change_cmode (ci->bot, chan, "+v", param);
	  }
	}
         /*************************** End Voice ********************/
         /*************************** De Voice *********************/
	else if (stricmp(cmd,"!devoice")==0) {
	   param = strtok(NULL, " ");

	   if (!is_chanop(source, chan) && (get_access(u, ci) < 4)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		if (!is_on_chan(source, chan))
			notice(ci->bot, source,"You are not in channel %s.", chan);
		else if (!is_voiced(source, chan))
			notice(ci->bot, source,"You already devoice in %s.", chan);
		else
		        change_cmode (ci->bot, chan, "-v", source);

	  } else {
		if (!is_on_chan(param, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if (!is_voiced(param, chan))
			notice(ci->bot, source,"User %s already devoice in %s.",param,chan);
		else
		        change_cmode (ci->bot, chan, "-v", param);
	  }
	}
         /*************************** End DeVoice ******************/
         /*************************** Halfop ***********************/
#ifdef DEFHALFOP
	else if (stricmp(cmd,"!halfop")==0) {
	   param = strtok(NULL, " ");

	   if (!is_chanop(source, chan) && (get_access(u, ci) < 5)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		if (!is_on_chan(source, chan))
			notice(ci->bot, source,"You are not in channel %s.", chan);
		else if (is_chanhop(source, chan))
			notice(ci->bot, source,"You already have halfop in %s.", chan);
		else
		        change_cmode (ci->bot, chan, "+h", source);

	  } else {
		User *u2 = finduser(param);
		if (!is_on_chan(param, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if ((ci->flags & CI_OPGUARD) && (get_access(u2, ci) < 4))
			notice(ci->bot, source,"Channel %s set opguard on please respect founder.",chan);
		else if (is_chanhop(param, chan))
			notice(ci->bot, source,"User %s already have halfop in %s.",param,chan);
		else
		        change_cmode (ci->bot, chan, "+h", param);
	  }
	}
         /*************************** End Halfop *******************/
	else if (stricmp(cmd,"!dehalfop")==0) {
	   param = strtok(NULL, " ");

	   if (!is_chanop(source, chan) && (get_access(u, ci) < 5)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		if (!is_on_chan(source, chan))
			notice(ci->bot, source,"You are not in channel %s.", chan);
		else if (!is_chanhop(source, chan))
			notice(ci->bot, source,"You already dehalfop in %s.", chan);
		else
		        change_cmode (ci->bot, chan, "-h", source);

	  } else {
		if (!is_on_chan(param, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if (!is_chanhop(param, chan))
			notice(ci->bot, source,"User %s already dehalfop in %s.",param,chan);
		else
		        change_cmode (ci->bot, chan, "-h", param);
	  }

	}
#endif
         /*************************** DeHalfop *********************/
         /*************************** Kick *************************/
	else if (stricmp(cmd,"!kick")==0) {
	   char *reason;
	   param = strtok(NULL, " ");
	   reason = strtok(NULL, " ");

	   if (!is_chanop(source, chan) && (get_access(u, ci) < 4)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		notice(ci->bot, source,"Syntax: !KICK <nick> [<reason>]");
	  } else {
		if (!is_on_chan(param, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if (
#ifdef DEFHALFOP
is_chanhop(param, chan) || 
#endif
is_chanop(param, chan) 
#ifdef DEFPROTECT
|| is_prots(param, chan)
#endif
)
			notice(ci->bot, source,"I don't kick op or protected user.");
		else {
		       char *av[3];
	               av[0] = sstrdup(chan);
	               av[1] = sstrdup(param);
	               if (!reason)
		          av[2] = sstrdup(buf);
	               else
	                  av[2] = sstrdup(reason);
	               send_cmd(ci->bot, "KICK %s %s :%s", av[0], av[1], av[2]);
	               do_kick(ci->bot, 3, av);
	               free(av[2]);
	               free(av[1]);
	               free(av[0]);
                }		
	  }
	}
         /************************** End Kick **********************/
         /************************** Ban ***************************/
	else if (stricmp(cmd,"!ban")==0) {
	   char *btype;
	   param = strtok(NULL, " ");
	   btype = strtok(NULL, " ");

	   if (!is_chanop(source, chan) && (get_access(u, ci) < 4)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		notice(ci->bot, source,"Syntax: !BAN <nick> ![<type>]");
	  } else {
		if (!is_on_chan(param, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if (
#ifdef DEFHALFOP
is_chanhop(param, chan) || 
#endif
is_chanop(param, chan) 
#ifdef DEFPROTECT
|| is_prots(param, chan)
#endif
)
			notice(ci->bot, source,"I don't ban op or protected user.");
		else {
		       User *u2 = finduser(param);
		       char *av[3];
	               av[0] = sstrdup(chan);
	               av[1] = sstrdup("+b");

		       if (btype && *btype == '!')
	               if (!btype || (stricmp(btype,"!1")==0))
		          av[2] = bcreate_mask(u2, 1);
	               else
	                  av[2] = bcreate_mask(u2, 2);
		       else
		          av[2] = bcreate_mask(u2, 1);

		       send_cmd(ci->bot, "MODE %s +b %s  %lu", chan, av[2], time(NULL));
		       do_cmode(s_ChanServ, 3, av);
	               free(av[2]);
	               free(av[1]);
	               free(av[0]);
                }		
	  }
	}
         /************************** End Ban ***********************/
         /************************** BKICK *************************/
	else if (stricmp(cmd,"!bkick")==0) {
	   char *reason, *btype, newreason[512];
	   param = strtok(NULL, " ");
	   btype = strtok(NULL, " ");
	   reason = strtok(NULL, "");

	   if (!is_chanop(source, chan) && (get_access(u, ci) < 4)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		notice(ci->bot, source,"Syntax: !BKICK <nick> ![<type>] [<reason>]");
	  } else {
		if (!is_on_chan(param, chan))
			notice(ci->bot, source,"User %s is not in %s.",param,chan);
		else if (
#ifdef DEFUNREAL
is_chanhop(param, chan) || 
#endif
is_chanop(param, chan)
#ifdef DEFPROTECT
|| is_prots(param, chan)
#endif
)
			notice(ci->bot, source,"I don't ban kick op or protected user.");
		else {
		       User *u2 = finduser(param);
		       char *av[3];
	               av[0] = sstrdup(chan);
	               av[1] = sstrdup("+b");

		       if (btype && *btype == '!')
	               if (!btype || (stricmp(btype,"!1")==0))
		          av[2] = bcreate_mask(u2, 1);
	               else
	                  av[2] = bcreate_mask(u2, 2);
		       else
		          av[2] = bcreate_mask(u2, 1);
			
		       send_cmd(ci->bot, "MODE %s +b %s  %lu", chan, av[2], time(NULL));
		       do_cmode(ci->bot, 3, av);

	               av[1] = sstrdup(param);

	               if (!reason)
		          av[2] = sstrdup("Ban kick requested");
	               else {
		          if (!(*btype == '!')) {
				  *newreason = 0;
				  strcat(newreason, btype);
				  strcat(newreason, " ");
				  strcat(newreason, reason);
		                  av[2] = sstrdup(newreason);
			   } else
		                  av[2] = sstrdup(reason);

		       }
	               send_cmd(ci->bot, "KICK %s %s :%s", av[0], av[1], av[2]);
	               do_kick(ci->bot, 3, av);

	               free(av[2]);
	               free(av[1]);
	               free(av[0]);
                }		
	  }
	}
         /************************** Ban Kick **********************/
         /************************** Un Ban ************************/
	else if (stricmp(cmd,"!unban")==0) {
	   char *btype;
	   param = strtok(NULL, " ");
	   btype = strtok(NULL, " ");

	   if (!is_chanop(source, chan) && (get_access(u, ci) < 4)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

	  if (!param) {
		notice(ci->bot, source,"Syntax: !UNBAN <nick>");
	  } else {
	       User *u2 = finduser(param);
	       Channel *c= findchan(chan);
	       char *av[3];
	       int i;
	       if (c && u2) {
	               av[0] = sstrdup(chan);
	               av[1] = sstrdup("-b");
		       for (i = c->bancount-1; i > -1; --i) {
		            if (match_usermask(c->bans[i], u2)) {
		                send_cmd(ci->bot, "MODE %s -b %s", chan, c->bans[i]);
		                av[2] = sstrdup(c->bans[i]);
		                do_cmode(s_ChanServ, 3, av);
		                free(av[2]);
		            }
		        }

	               free(av[1]);
        	       free(av[0]);
		}
	  }
	}
         /************************** End Unban *********************/
         /************************** HELP **************************/
	else if (stricmp(cmd,"!help")==0) {

	   if (!is_chanop(source, chan) && (get_access(u, ci) < 4)) {
		notice(ci->bot, source,"You don't have access.");
		return;
	   }

		if (!is_on_chan(source, chan))
			notice(ci->bot, source,"You are not in channel %s.", chan);
		else {
		notice(ci->bot, source,"!op       \2[<nick>]\2");
		notice(ci->bot, source,"!deop     \2[<nick>]\2");
		notice(ci->bot, source,"!voice    \2[<nick>]\2");
		notice(ci->bot, source,"!devoice  \2[<nick>]\2");
#ifdef DEFHALFOP
		notice(ci->bot, source,"!halfop   \2[<nick>]\2");
		notice(ci->bot, source,"!dehalfop \2[<nick>]\2");
#endif
		notice(ci->bot, source,"!kick     \2<nick>\2 \2[<reason>]\2");
		notice(ci->bot, source,"!ban      \2<nick>\2 \2![<type = 1|2>]\2");
		notice(ci->bot, source,"!bkick    \2<nick>\2 \2![<type = 1|2>]\2 \2[<reason>]\2");
		notice(ci->bot, source,"!GWord    \2word or char\2");
		notice(ci->bot, source,"!GNum     \2number\2");
		notice(ci->bot, source,"End of bot command.");
	}
       }
         /************************** End Help **********************/
   } else {
   }
}

/******************************************************************/

/**********************************************************************/
int sockcheck2(char *host)
{
  struct hostent *hp;
  struct sockaddr_in sa;
  int sock = -1;
  unsigned char packet[12];

  if (!(hp = gethostbyname(host))) return;


  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) return;

  bzero(&sa, sizeof(sa));
  sa.sin_family = AF_INET;
  sa.sin_port = htons (SOCKS5PORT);
  bcopy(hp->h_addr, (char *) &sa.sin_addr, hp->h_length);
  if (connect(sock, (struct sockaddr *) &sa, sizeof (sa)) == 0) {
    packet[0] = 5;
    packet[1] = 1;
    packet[2] = 0;

    if (debug)
        log("Checking for Unsecured socks5 for %s",host);
    if (write(sock, packet, 3) == 3) {
      bzero(packet, 2);
      if (read(sock, packet, 2) == 2 && packet[0]==5 && packet[1]==0) {
        close(sock);
	send_cmd(s_Agent, "GLINE *@%s 1800 :Unsecured socks5", host);
        return 1;
      }
    }
    close(sock);

    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) return;

    bzero(&sa, sizeof(sa));
    sa.sin_family = AF_INET;
    sa.sin_port = htons (SOCKS4PORT);
    bcopy(hp->h_addr, (char *) &sa.sin_addr, hp->h_length);
    if (debug)
        log("Checking for Unsecured socks4 for %s",host);
    if (connect(sock, (struct sockaddr *) &sa, sizeof (sa)) == 0) {
      unsigned long ip = htonl(sa.sin_addr.s_addr);
      packet[0] = 4;
      packet[1] = 1;
      packet[2] = (((unsigned short) SOCKS4PORT) >> 8);
      packet[3] = (((unsigned short) SOCKS4PORT) & 255);
      packet[4] = (ip >> 24);
      packet[5] = ((ip >> 16) & 255);
      packet[6] = ((ip >> 8) & 255);
      packet[7] = (ip & 255);
      packet[8] = 0;
      if (write(sock, packet, 9) == 9) {
        bzero(packet, 9);
        if (read(sock, packet, 9)>=4 && packet[1] == 90) {
          close(sock);
	  send_cmd(s_Agent, "GLINE *@%s 1800 :Unsecured socks4", host);
          return 1;
        }
      }
    }
  }
  close(sock);


  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) return;
  bzero(&sa, sizeof(sa));
  sa.sin_family = AF_INET;
  sa.sin_port = htons(6667);
  bcopy(hp->h_addr, (char *) &sa.sin_addr, hp->h_length);
  if (connect(sock, (struct sockaddr *) &sa, sizeof (sa)) == 0) {
    memset(packet, 0, sizeof(packet));
    if (read(sock, packet, 10)) {
        wallops(s_Agent, "Possible an IRC Server detected from: \2%s\2", host);
        close(sock);
    }
  }
  close(sock);

  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) return;

  bzero(&sa, sizeof(sa));
  sa.sin_family = AF_INET;
  sa.sin_port = htons(WINGATE);
  bcopy(hp->h_addr, (char *) &sa.sin_addr, hp->h_length);

  if (debug)
        log("Checking for Unsecured wingate for %s",host);

  if (connect(sock, (struct sockaddr *) &sa, sizeof (sa)) == 0) {
    memset(packet, 0, sizeof(packet));
    if (read(sock, packet, 8)==8
        && (!strncasecmp(packet, "WinGate>", 8)
            || !strncasecmp(packet, "Too many", 8))) {
        close(sock);
	send_cmd(s_Agent, "GLINE *@%s 1800 :Unsecured wingate", host);
      return 1;
    }
  }
  close(sock);
  return 0;
}
/************************************************************************/

